"""This module stores API defaults."""

from enum import IntEnum
from pathlib import Path
from . import __version__

try:
    from sys import _MEIPASS
except ImportError:
    _MEIPASS = None

__all__ = [
    'Limits',
    'Scrypt',
    'VERSION',
    'VERBYTE',
    'DEF_TGBOX_NAME',
    'REMOTEBOX_PREFIX',
    'DEF_NO_FOLDER',
    'DEF_UNK_FOLDER',
    'PREFIX',
    'BOX_IMAGE_PATH',
    'WORDS_PATH',
    'FFMPEG',
    'ABSPATH',
    'DOWNLOAD_PATH',
    'PYINSTALLER_DATA'
]
class Limits(IntEnum):
    """Default TGBOX API limits"""
    # We store metadata size in three bytes, but
    # by default it's size is limited to 1MB. You
    # can set it up to the 256^3-1, but to share
    # your files with other people they should have
    # a METADATA_MAX that >= than yours.
    METADATA_MAX: int=1000000
    # Max FILE_PATH length on Linux
    # is 4096 bytes (4KiB).
    FILE_PATH_MAX: int=4096

class Scrypt(IntEnum):
    """Default Scrypt KDF configuration"""
    # See https://en.wikipedia.org/wiki/Scrypt for base info about Scrypt
    # -------------------------------------------------------------------
    # This salt affects basekeys, you can change it to protect your RemoteBox
    # from bruteforcing, but be sure to backup your own salt, because if you
    # lose it, then it will be impossible to recover your decryption key for Tgbox.
    # Change this isn't necessary at all, because if you use *strong* password or
    # generated keys.Phrase, then it's already should be impossible to brute force.
    # Think of it as 2FA. Changed salt doesn't protect you if you leaked your mainkey,
    # but will protect you from passphrase leakage.
    SALT: int=0x37CE65C834C6EFE05DFAD02413C0950072A1FE3ED48A33368333848D9C782167
    # You can change any Scrypt params. Please note that by default resulted
    # key will be hashed with sha256, so BaseKey is always 32-byte.
    # Default is balanced to use 1GB of RAM.
    DKLEN: int=32
    N:     int=2**20
    R:     int=8
    P:     int=1

# Path that will be used to save downloaded files.
DOWNLOAD_PATH: Path=Path('DownloadsTGBOX')

VERSION: str=__version__
VERBYTE: bytes=b'\x01'

DEF_TGBOX_NAME:   str='TGBOX'
REMOTEBOX_PREFIX: str=f'{DEF_TGBOX_NAME}[{VERBYTE.hex()}]: '

DEF_NO_FOLDER:  Path=Path('NO_FOLDER')
DEF_UNK_FOLDER: Path=Path('UNKNOWN_FOLDER')

PREFIX: bytes=b'\x00TGBOX'

ABSPATH: Path = Path(_MEIPASS) if _MEIPASS is not None \
    else Path(__file__).parent

# Get path to "other" folder where we store
# words.txt and tgbox_logo.png files.
_other: Path = ABSPATH / 'other'

# We will use it in subprocess.call
FFMPEG = 'ffmpeg'

# You can add ffmpeg.exe to 'other' folder
# before build with PyInstaller on Windows or
# just if you want TGBOX to make file thumbnails
# or extract duration. It will be added to
# your resulted executable.
#
# https://www.ffmpeg.org/download.html#build-windows
#
for file in _other.iterdir():
    if file.name == 'ffmpeg.exe':
        FFMPEG = _other / 'ffmpeg.exe'

# By default, PyInstaller will not grab files
# from 'other' folder. To resolve this error
# you will need to manually add it to .spec file.
#
# See
#   '.spec file' example:
#        github.com/NotStatilko/tgbox-cli/blob/main/tgbox_cli.spec
#
#   'Installation' in TGBOX docs:
#       tgbox.readthedocs.io/en/latest/installation.html
#
PYINSTALLER_DATA: dict = {
    str(Path('other', i.name)): str(i)
    for i in _other.glob('*')
}
BOX_IMAGE_PATH: Path = _other / 'tgbox_logo.png'
WORDS_PATH: Path = _other / 'words.txt'



"""This module stores wrappers around Tgbox SQL DB."""

from typing import (
    Optional, Union,
    AsyncGenerator
)
from os import PathLike
from pathlib import Path

import aiosqlite

from ..defaults import (
    Limits, DOWNLOAD_PATH,
    DEF_NO_FOLDER, DEF_UNK_FOLDER,
)
from ..errors import PathIsDirectory
from ..tools import anext

__all__ = ['SqlTableWrapper', 'TgboxDB', 'TABLES']

TABLES = {
    'BOX_DATA': (
        ('BOX_CHANNEL_ID', 'BLOB NOT NULL'),
        ('BOX_CR_TIME', 'BLOB NOT NULL'),
        ('BOX_SALT', 'BLOB NOT NULL'),
        ('MAINKEY', 'BLOB'),
        ('SESSION', 'BLOB NOT NULL'),
        ('API_ID', 'BLOB NOT NULL'),
        ('API_HASH', 'BLOB NOT NULL'),
    ),
    'FILES': (
        ('ID', 'INTEGER PRIMARY KEY'),
        ('UPLOAD_TIME', 'BLOB NOT NULL'),
        ('PPATH_HEAD', 'BLOB NOT NULL'),
        ('FILEKEY', 'BLOB'),
        ('FINGERPRINT', 'BLOB'),
        ('METADATA', 'BLOB NOT NULL'),
        ('UPDATED_METADATA', 'BLOB')
    ),
    'PATH_PARTS': (
        ('ENC_PART', 'BLOB NOT NULL'),
        ('PART_ID', 'BLOB NOT NULL'),
        ('PARENT_PART_ID', 'BLOB'),
    ),
    'DEFAULTS': (                            # Default value
        ('METADATA_MAX', 'INTEGER NOT NULL', int(Limits.METADATA_MAX)),
        ('FILE_PATH_MAX', 'INTEGER NOT NULL', int(Limits.FILE_PATH_MAX)),

        ('DOWNLOAD_PATH', 'TEXT NOT NULL', str(DOWNLOAD_PATH)),
        ('DEF_NO_FOLDER', 'TEXT NOT NULL', str(DEF_NO_FOLDER)),
        ('DEF_UNK_FOLDER', 'TEXT NOT NULL', str(DEF_UNK_FOLDER))
    )
}
class SqlTableWrapper:
    """A low-level wrapper to SQLite Tables."""
    def __init__(self, aiosql_conn, table_name: str):
        self._table_name = table_name
        self._aiosql_conn = aiosql_conn

    async def __aiter__(self) -> tuple:
        """Will yield rows as self.select without ``sql_statement``"""
        async for row in self.select():
            yield row

    @property
    def table_name(self) -> str:
        """Returns table name"""
        return self._table_name

    async def count_rows(self) -> int:
        """Execute ``SELECT count(*) from TABLE_NAME``"""
        cursor = await self._aiosql_conn.execute(
            f'SELECT count(*) FROM {self._table_name}'
        )
        return (await cursor.fetchone())[0]

    # TODO: Auto add FROM <TABLE_NAME> for select
    #       and INSERT INTO <TABLE_NAME> for insert
    async def select(self, sql_tuple: Optional[tuple] = None) -> AsyncGenerator:
        """
        If ``sql_tuple`` isn't specified, then will be used
        ``(SELECT * FROM TABLE_NAME, ())`` statement.
        """
        if not sql_tuple:
            sql_tuple = (f'SELECT * FROM {self._table_name}',())

        cursor = await self._aiosql_conn.execute(*sql_tuple)
        async for row in cursor: yield row

    async def select_once(self, sql_tuple: Optional[tuple] = None) -> tuple:
        """
        Will return first row which match the ``sql_tuple``,
        see ``select()`` method for ``sql_tuple`` details.
        """
        return await anext(self.select(sql_tuple=sql_tuple))

    async def insert(
            self, *args, sql_statement: Optional[str] = None,
            commit: bool=True) -> None:
        """
        If ``sql_statement`` isn't specified, then will be used
        ``INSERT INTO TABLE_NAME values (...)``.

        This method doesn't check if you insert correct data
        or correct amount of it, you should know DB structure.
        """
        if not sql_statement:
            sql_statement = (
                f'INSERT INTO {self._table_name} values ('
                + ('?,' * len(args))[:-1] + ')'
            )
        await self._aiosql_conn.execute(sql_statement, args)
        if commit: await self._aiosql_conn.commit()

    async def execute(self, sql_tuple: tuple, commit: bool=True):
        result = await self._aiosql_conn.execute(*sql_tuple)
        if commit: await self._aiosql_conn.commit()
        return result # Returns Cursor object

    async def commit(self) -> None:
        await self._aiosql_conn.commit()

class TgboxDB:
    def __init__(self, db_path: Union[PathLike, str]):
        """
        Arguments:
            db_path (``PathLike``, ``str``):
                Path to the Tgbox DB.
        """
        if isinstance(db_path, PathLike):
            self._db_path = db_path
        else:
            self._db_path = Path(db_path)

        self._aiosql_db = None
        self._aiosql_db_is_closed = None
        self._name = self._db_path.name

        if self._db_path.is_dir():
            raise PathIsDirectory('Path is directory.')

    @property
    def name(self) -> str:
        return self._name

    @property
    def db_path(self) -> PathLike:
        return self._db_path

    @property
    def closed(self) -> bool:
        """
        This method will return ``None`` if DB wasn't opened,
        False if it's still opened, True if it's was closed.
        """
        return self._aiosql_db_is_closed

    @staticmethod
    async def create(db_path: Union[str, PathLike]) -> 'TgboxDB':
        return await TgboxDB(db_path).init()

    async def close(self) -> None:
        await self._aiosql_db.close()
        self._aiosql_db_is_closed = True

    async def init(self) -> 'TgboxDB':
        self._aiosql_db = await aiosqlite.connect(self._db_path)

        for table, data in TABLES.items():
            try:
                columns = ', '.join((f'{i[0]} {i[1]}' for i in data))
                await self._aiosql_db.execute(
                    f'CREATE TABLE {table} ({columns})'
                )
                if table == 'DEFAULTS':
                    sql_statement = (
                        f'INSERT INTO {table} VALUES ('
                        + ('?,' * len(data))[:-1] + ')'
                    )
                    await self._aiosql_db.execute(
                        sql_statement, [i[2] for i in data]
                    )
            except aiosqlite.OperationalError: # Table exists
                table_columns = await self._aiosql_db.execute(
                    f'PRAGMA table_info({table})'
                )
                table_columns = set((i[1] for i in await table_columns.fetchall()))
                required_columns = set((i[0] for i in data))

                if table_columns != required_columns:
                    table_columns &= required_columns

                    await self._aiosql_db.execute(
                        f'CREATE TABLE "updated!{table}" ({columns})'
                    )
                    table_columns_str = ', '.join(table_columns)
                    await self._aiosql_db.execute(
                        f"""INSERT INTO "updated!{table}" ({table_columns_str}) """
                        f"""SELECT {table_columns_str} FROM {table}"""
                    )
                    await self._aiosql_db.execute(f'DROP TABLE {table}')
                    await self._aiosql_db.execute(
                        f'ALTER TABLE "updated!{table}" RENAME TO {table}'
                    )
        await self._aiosql_db.commit()
        self._aiosql_db_is_closed = False

        tables = await self._aiosql_db.execute(
            "SELECT name FROM sqlite_master WHERE type='table'"
        )
        for table in (await tables.fetchall()):
            setattr(self, table[0], SqlTableWrapper(self._aiosql_db, table[0]))

        return self



"""Module with all API utils."""

from typing import (
    BinaryIO, Optional,
    Union, AsyncGenerator
)
from os import PathLike
from dataclasses import dataclass
from base64 import urlsafe_b64encode

from telethon.tl.custom.file import File
from telethon.sessions import StringSession

from telethon.tl.types import Photo, Document
from telethon.tl.types.auth import SentCode

from telethon import TelegramClient as TTelegramClient
from telethon.errors import SessionPasswordNeededError
from telethon.tl.functions.auth import ResendCodeRequest

try:
    from regex import search as re_search
except ImportError:
    from re import search as re_search

from ..defaults import VERSION
from ..tools import anext, SearchFilter
from ..fastelethon import download_file

from .db import TABLES, TgboxDB

__all__ = [
    'search_generator',
    'DirectoryRoot',
    'PreparedFile',
    'TelegramClient',
    'DefaultsTableWrapper',
    'RemoteBoxDefaults'
]
class TelegramClient(TTelegramClient):
    """
    A little extend to the ``telethon.TelegramClient``.

    This class inherits Telethon's TelegramClient and support
    all features that has ``telethon.TelegramClient``.

    Typical usage:

    .. code-block:: python

        from asyncio import run as asyncio_run
        from tgbox.api import TelegramClient, make_remotebox
        from getpass import getpass # For hidden input

        PHONE_NUMBER = '+10000000000' # Your phone number
        API_ID = 1234567 # Your API_ID: https://my.telegram.org
        API_HASH = '00000000000000000000000000000000' # Your API_HASH

        async def main():
            tc = TelegramClient(
                phone_number = PHONE_NUMBER,
                api_id = API_ID,
                api_hash = API_HASH
            )
            await tc.connect()
            await tc.send_code()

            await tc.log_in(
                code = int(input('Code: ')),
                password = getpass('Pass: ')
            )
            erb = await make_remotebox(tc)

        asyncio_run(main())
    """
    def __init__(
            self, api_id: int, api_hash: str,
            phone_number: Optional[str] = None,
            session: Optional[Union[str, StringSession]] = None,
            **kwargs) -> None:
        """
        .. note::
            You should specify at least ``session`` or ``phone_number``.

        Arguments:
            api_id (``int``):
                API_ID from https://my.telegram.org.

            api_hash (``int``):
                API_HASH from https://my.telegram.org.

            phone_number (``str``, optional):
                Phone number linked to your Telegram
                account. You may want to specify it
                to recieve log-in code. You should
                specify it if ``session`` is ``None``.

            session (``str``, ``StringSession``, optional):
                ``StringSession`` that give access to
                your Telegram account. You can get it
                after connecting and signing in via
                ``TelegramClient.session.save()`` method.

        ..tip::
            This ``TelegramClient`` support all keywoard
            arguments (**kwargs) that support parent
            ``telethon.TelegramClient`` object.
        """
        if not session and not phone_number:
            raise ValueError(
                'You should specify at least ``session`` or ``phone_number``.'
            )
        super().__init__(
            StringSession(session),
            api_id, api_hash, **kwargs
        )
        self.__version__ = VERSION
        self._api_id, self._api_hash = api_id, api_hash
        self._phone_number = phone_number

    async def send_code(self, force_sms: Optional[bool]=False) -> SentCode:
        """
        Sends the Telegram code needed to login to the given phone number.

        Arguments:
            force_sms (``bool``, optional):
                Whether to force sending as SMS.
        """
        return await self.send_code_request(
            self._phone_number, force_sms=force_sms
        )
    async def log_in(
            self, password: Optional[str] = None,
            code: Optional[Union[int,str]] = None) -> None:
        """
        Logs in to Telegram to an existing user account.
        You should only use this if you are not signed in yet.

        Arguments:
            password (``str``, optional):
                Your 2FA password. You can ignore
                this if you don't enabled it yet.

            code (``int``, optional):
                The code that Telegram sent you after calling
                ``TelegramClient.send_code()`` method.
        """
        if not await self.is_user_authorized():
            try:
                await self.sign_in(self._phone_number, code)
            except SessionPasswordNeededError:
                await self.sign_in(password=password)

    async def resend_code(self, sent_code: SentCode) -> SentCode:
        """
        Will send you login code again. This can be used to
        force Telegram send you SMS or Call to dictate code.

        Arguments:
            sent_code (``SentCode``):
                Result of the ``tc.send_code`` or
                result of the ``tc.resend_code`` method.

        Example:

        .. code-block:: python

            tc = tgbox.api.TelegramClient(...)
            sent_code = await tc.send_code()
            sent_code = await tc.resend_code(sent_code)
        """
        return await self(ResendCodeRequest(
            self._phone_number, sent_code.phone_code_hash)
        )
@dataclass
class PreparedFile:
    """
    This dataclass store data needed for upload
    by ``DecryptedRemoteBox.push_file`` in future.

    Usually it's only for internal use.
    """
    dlb: 'tgbox.api.local.DecryptedLocalBox'
    file: BinaryIO
    filekey: 'tgbox.keys.FileKey'
    filesize: int
    filepath: PathLike
    filesalt: bytes
    fingerprint: bytes
    metadata: bytes
    imported: bool

    def set_file_id(self, id: int):
        """You should set ID after pushing to remote"""
        self.file_id = id

    def set_upload_time(self, upload_time: int):
        """You should set time after pushing to remote"""
        self.upload_time = upload_time

class DirectoryRoot:
    """
    Type used to specify that you want to
    access absolute local directory root.

    This class doesn't have any methods,
    please use it only for ``lbd.iterdir``
    """

# TODO: Improve SearchFilter
async def search_generator(
        sf: SearchFilter, it_messages: Optional[AsyncGenerator] = None,
        lb: Optional['tgbox.api.local.DecryptedLocalBox'] = None,
        cache_preview: bool = True) -> AsyncGenerator:
    """
    Generator used to search for files in dlb and rb. It's
    only for internal use, and you shouldn't use it in your
    own projects.

    If file is exported from other RemoteBox and was imported to your
    LocalBox, then you can specify ``dlb`` as ``lb``. AsyncGenerator
    will try to get ``FileKey`` and decrypt ``EncryptedRemoteBoxFile``.
    Otherwise imported file will be ignored.
    """
    in_func = re_search if sf.in_filters['re'] else lambda p,s: p in s

    if it_messages:
        iter_from = it_messages
    else:
        min_id = sf.in_filters['min_id'][-1] if sf.in_filters['min_id'] else None
        max_id = sf.in_filters['max_id'][-1] if sf.in_filters['max_id'] else None
        iter_from = lb.files(min_id=min_id, max_id=max_id, cache_preview=cache_preview)

    if not iter_from:
        raise ValueError('At least it_messages or lb must be specified.')

    async for file in iter_from:
        if hasattr(file, '_message'): # *RemoteBoxFile
            file_size = file.file_size
            file_path = file.file_path if file.file_path else ''
        elif hasattr(file, '_tgbox_db'): # *LocalBoxFile
            file_size = file.size
            await file.directory.lload(full=True)
            file_path = str(file.directory)
        else:
            continue

        # We will use it as flags, the first
        # is for 'include', the second is for
        # 'exclude'. Both should be True to
        # match SearchFilter filters.
        yield_result = [True, True]

        for indx, filter in enumerate((sf.in_filters, sf.ex_filters)):
            if filter['imported']:
                if bool(file.imported) != bool(filter['imported']):
                    if indx == 0: # O is Include
                        yield_result[indx] = False
                        break

                elif bool(file.imported) == bool(filter['imported']):
                    if indx == 1: # 1 is Exclude
                        yield_result[indx] = False
                        break

            for mime in filter['mime']:
                if in_func(mime, file.mime):
                    if indx == 1:
                        yield_result[indx] = False
                    break
            else:
                if filter['mime']:
                    if indx == 0:
                        yield_result[indx] = False
                        break

            if filter['min_time']:
                if file.upload_time < filter['min_time'][-1]:
                    if indx == 0:
                        yield_result[indx] = False
                        break

                elif file.upload_time >= filter['min_time'][-1]:
                    if indx == 1:
                        yield_result[indx] = False
                        break

            if filter['max_time']:
                if file.upload_time > filter['max_time'][-1]:
                    if indx == 0:
                        yield_result[indx] = False
                        break

                elif file.upload_time <= filter['max_time'][-1]:
                    if indx == 1:
                        yield_result[indx] = False
                        break

            if filter['min_size']:
                if file_size < filter['min_size'][-1]:
                    if indx == 0:
                        yield_result[indx] = False
                        break

                elif file_size >= filter['min_size'][-1]:
                    if indx == 1:
                        yield_result[indx] = False
                        break

            if filter['max_size']:
                if file_size > filter['max_size'][-1]:
                    if indx == 0:
                        yield_result[indx] = False
                        break

                elif file_size <= filter['max_size'][-1]:
                    if indx == 1:
                        yield_result[indx] = False
                        break

            if filter['min_id']:
                if file.id < filter['min_id'][-1]:
                    if indx == 0:
                        yield_result[indx] = False
                        break

                elif file.id >= filter['min_id'][-1]:
                    if indx == 1:
                        yield_result[indx] = False
                        break

            if filter['max_id']:
                if file.id > filter['max_id'][-1]:
                    if indx == 0:
                        yield_result[indx] = False
                        break

                elif file.id <= filter['max_id'][-1]:
                    if indx == 1:
                        yield_result[indx] = False
                        break

            for id in filter['id']:
                if file.id == id:
                    if indx == 1:
                        yield_result[indx] = False
                    break
            else:
                if filter['id']:
                    if indx == 0:
                        yield_result[indx] = False
                        break

            if hasattr(file, '_cattrs'):
                for cattr in filter['cattrs']:
                    for k,v in cattr.items():
                        if k in file.cattrs:
                            if in_func(v, file.cattrs[k]):
                                if indx == 1:
                                    yield_result[indx] = False
                                break
                    else:
                        if filter['cattrs']:
                            if indx == 0:
                                yield_result[indx] = False
                                break

            for filter_file_path in filter['file_path']:
                if in_func(str(filter_file_path), file_path):
                    if indx == 1:
                        yield_result[indx] = False
                    break
            else:
                if filter['file_path']:
                    if indx == 0:
                        yield_result[indx] = False
                        break

            for file_name in filter['file_name']:
                if in_func(file_name, file.file_name):
                    if indx == 1:
                        yield_result[indx] = False
                    break
            else:
                if filter['file_name']:
                    if indx == 0:
                        yield_result[indx] = False
                        break

            for file_salt in filter['file_salt']:
                if isinstance(file_salt, str):
                    fsalt = urlsafe_b64encode(file.file_salt).decode()
                else:
                    fsalt = file.file_salt

                if in_func(file_salt, fsalt):
                    if indx == 1:
                        yield_result[indx] = False
                    break
            else:
                if filter['file_salt']:
                    if indx == 0:
                        yield_result[indx] = False
                        break

            for verbyte in filter['verbyte']:
                if verbyte == file.verbyte:
                    if indx == 1:
                        yield_result[indx] = False
                    break
            else:
                if filter['verbyte']:
                    if indx == 0:
                        yield_result[indx] = False
                        break

        if all(yield_result):
            yield file
        else:
            continue

class _TelegramVirtualFile:
    """
    We use this class for re-upload to RemoteBox
    files that already was uploaded to any other
    Telegram chat. That's only for internal use.
    """
    def __init__(
            self, document: Union[Photo, Document],
            session: StringSession,
            api_id: int, api_hash: str):

        self.downloader = None
        self.document = document

        self.tc = TelegramClient(
            session=session,
            api_id=api_id,
            api_hash=api_hash
        )
        self.tc = self.tc.connect()
        self._client_initialized = False

        file = File(document)

        self.name = file.name
        self.size = file.size
        self.mime = file.mime_type

        self.duration = file.duration\
            if file.duration else 0

    async def get_preview(self, quality: int=1) -> bytes:
        if not self._client_initialized:
            self.tc = await self.tc # connect
            self._client_initialized = True

        if hasattr(self.document,'sizes')\
            and not self.document.sizes:
                return b''

        if hasattr(self.document,'thumbs')\
            and not self.document.thumbs:
                return b''

        return await self.tc.download_media(
            message = self.document,
            thumb = quality, file = bytes
        )
    async def read(self, size: int) -> bytes:
        if not self._client_initialized:
            self.tc = await self.tc
            self._client_initialized = True

        if not self.downloader:
            self.downloader = download_file(
                self.tc, self.document
            )
        chunk = await anext(self.downloader)
        return chunk

class DefaultsTableWrapper:
    """
    This little class will wrap around the
    DEFAULTS table of TGBOX DB and will
    fetch all contents of it.

    You can await the ``change`` coroutine
    to change default values to your own.
    """
    def __init__(self, tgbox_db: TgboxDB):
        """
        Arguments:
            tgbox_db (``TgboxDB``):
                An initialized ``TgboxDB``.
        """
        self._tgbox_db = tgbox_db
        self._initialized = False

    @property
    def initialized(self) -> bool:
        return self._initialized

    async def init(self) -> 'DefaultsTableWrapper':
        """Fetch the defaults and initialize"""
        if self._tgbox_db.closed:
            await self._tgbox_db.init()

        defaults = await self._tgbox_db.DEFAULTS.select_once()
        for default, value in zip(TABLES['DEFAULTS'], defaults):
            setattr(self, default[0], value)

        self._initialized = True
        return self

    async def change(self, key: str, value) -> None:
        """
        This method can change the defaults values

        Arguments:
            key (``str``):
                Key to change, i.e METADATA_MAX.

            value:
                Key's new value.

        .. warning::
            We **don't** verify here that value
            type corresponds to real type of Key
            or that value doesn't overflow the
            allowed value maximum. Be sure to
            specify the correct Key values.

        Example:

        .. code-block:: python

            from asyncio import run as asyncio_run

            from tgbox.defaults import DEF_TGBOX_NAME
            from tgbox.api.db import TgboxDB
            from tgbox.api.utils import DefaultsTableWrapper

            async def main():
                # Make a DefaultsTableWrapper object
                tdb = await TgboxDB(DEF_TGBOX_NAME).init()
                dtw = await DefaultsTableWrapper(tdb).init()

                # Change METADATA_MAX to the max allowed size
                dtw.change('METADATA_MAX', 256**3-1)

                # Access DTW from the DecryptedLocalBox
                ... # Some code was omited here
                # Change the default download path
                dlb.defaults.change('DOWNLOAD_PATH', 'Downloads')

            asyncio_run(main())
        """
        getattr(self, key) # Vetrify that Key exist
        await self._tgbox_db.DEFAULTS.execute((
            f'UPDATE DEFAULTS SET {key}=?', (value,)
        ))

@dataclass
class RemoteBoxDefaults:
    METADATA_MAX: int
    FILE_PATH_MAX: int
    DEF_UNK_FOLDER: Union[str, PathLike]
    DEF_NO_FOLDER: Union[str, PathLike]
    DOWNLOAD_PATH: Union[str, PathLike]



"""This module stores utils required by API."""

try:
    from regex import search as re_search
except ImportError:
    from re import search as re_search

from asyncio import (
    iscoroutine, get_event_loop
)
from copy import deepcopy
from pprint import pformat
from hashlib import sha256
from random import randrange

from subprocess import PIPE, run as subprocess_run
from typing import BinaryIO, Optional, Dict, Generator

from io import BytesIO
from os import PathLike
from pathlib import Path
from functools import partial
from os import remove as remove_file

from .errors import (
    ConcatError,
    PreviewImpossible,
    DurationImpossible
)
from .defaults import FFMPEG
from .keys import FileKey, MainKey
from .crypto import AESwState as AES

__all__ = [
    'prbg', 'anext',
    'SearchFilter',
    'OpenPretender',
    'PackedAttributes',
    'int_to_bytes',
    'bytes_to_int',
    'get_media_duration',
    'make_media_preview',
    'ppart_id_generator'
]
anext = lambda agen: agen.__anext__()

class _TypeList:
    """
    This is cutted version of ``list()`` that
    checks type on ``.append(...)``.

    You can specify multiply types with
    ``tuple``, e.g: ``_TypeList((int, float))``

    * The list will try to change value type if \
      ``isinstance(value, type_) is False`` to \
      ``value = type_(value)``. Otherwise ``TypeError``.
    """
    def __init__(self, type_, *args):
        self.type = type_ if isinstance(type_, tuple) else (type_,)
        self.list = [self.__check_type(i) for i in args]

    def __bool__(self):
        return bool(self.list)

    def __iter__(self):
        for i in self.list:
            yield i

    def __getitem__(self, sl):
        return self.list[sl]

    def __repr__(self) -> str:
        return f'_TypeList({self.type}, *{self.list})'

    def __check_type(self, value):
        for type_ in self.type:
            if isinstance(value, type_):
                return value
        else:
            for type_ in self.type:
                try:
                    if isinstance(b'', type_):
                        # bytes(str) doesn't work
                        return type_(value,'utf-8')
                    else:
                        return type_(value)
                except:
                    pass
            else:
                raise TypeError(
                    f'Invalid type! Expected {self.type}, got {type(value)}'
                )
    def append(self, value):
        self.list.append(self.__check_type(value))

    def extend(self, list):
        self.list.extend([self.__check_type(i) for i in list])

    def clear(self):
        self.list.clear()

class SearchFilter:
    """
    Container that filters search in ``DecryptedRemoteBox`` or
    ``DecryptedLocalBox``.

    The ``SearchFilter`` has **two** filters: the **Include**
    and **Exclude**. On search, all matching to **include**
    files will be **yielded**, while all matching to
    **exclude** will be **not yielded** (ignored).

    * The ``tgbox.tools.search_generator`` will firstly check for \
      **include** filters, so its priority higher.

    * The ``tgbox.tools.search_generator`` will yield files that \
      match **all** of filters, **not** one of it.

    * The ``SearchFilter`` accepts ``list`` as kwargs \
      value. You can ``SearchFilter(id=[3,5,10])``.

    * The ``SearchFilter(**kwargs)`` will add all filters \
      to the **include**. Also use ``SearchFilter.include(...)`` \
      & ``SearchFilter.exclude(...)`` methods after initializion.

    All filters:
        * **id** *integer*: File ID

        * **cattrs** *dict*: File CustomAttributes:
            To search for CATTRS you need to specify a dict.

            E.g: If *file* ``cattrs={b'comment': b'hi!'}``, then
            *filter* ``cattrs={b'comment': b'h'}`` will match.

            By default, ``tgbox.tools.search_generator`` will
            use an ``in``, like ``b'h' in b'hi!'``, but you
            can set a ``re`` flag to use regular expressions,
            so *filter* ``cattrs={b'comment': b'hi(.)'}`` will match.

        * **file_path** *pathlib.Path*, *str*
        * **file_name** *bytes*: File name
        * **file_salt** *bytes*: File salt
        * **verbyte**   *bytes*: File version byte

        * **min_id** *integer*: File ID should be > min_id
        * **max_id** *integer*: File ID should be < max_id

        * **min_size** *integer*: File Size should be > min_size
        * **max_size** *integer*: File Size should be < max_size

        * **min_time** *integer/float*: Upload Time should be > min_time
        * **max_time** *integer/float*: Upload Time should be < max_time

        * **mime**     *str*:  File mime type
        * **imported** *bool*: Yield only imported files
        * **re**       *bool*: re_search for every ``bytes``
    """
    def __init__(self, **kwargs):
        self.in_filters = {
            'cattrs':    _TypeList(dict),
            'file_path': _TypeList(str),
            'file_name': _TypeList(str),
            'file_salt': _TypeList((bytes,str)),
            'verbyte':   _TypeList(bytes),
            'id':        _TypeList(int),
            'min_id':    _TypeList(int),
            'max_id':    _TypeList(int),
            'min_size':  _TypeList(int),
            'max_size':  _TypeList(int),
            'min_time':  _TypeList((int,float)),
            'max_time':  _TypeList((int,float)),
            'mime':      _TypeList(str),
            'imported':  _TypeList(bool),
            're':        _TypeList(bool),
        }
        self.ex_filters = deepcopy(self.in_filters)
        self.include(**kwargs)

    def __repr__(self) -> str:
        return pformat({
            'include': self.in_filters,
            'exclude': self.ex_filters
        })
    def include(self, **kwargs) -> 'SearchFilter':
        """Will extend included filters"""
        for k,v in kwargs.items():
            if isinstance(v, list):
                self.in_filters[k].extend(v)
            else:
                self.in_filters[k].append(v)
        return self

    def exclude(self, **kwargs) -> 'SearchFilter':
        """Will extend excluded filters"""
        for k,v in kwargs.items():
            if isinstance(v, list):
                self.ex_filters[k].extend(v)
            else:
                self.ex_filters[k].append(v)
        return self

class PackedAttributes:
    """
    This class is used to pack items to
    bytestring. We use it to pack file
    metadata as well as pack a user's
    custom attributes (cattrs), then
    encrypt it and attach to RBFile.

    We store key/value length in 3 bytes,
    so the max key/value length is 256^3-1.

    <key-length>key<value-length>value<...>
    """
    @staticmethod
    def pack(**kwargs) -> bytes:
        """
        Will make bytestring from your kwargs.
        Any kwarg **always** must be ``bytes``.

        ``make(x=5)`` will not work;
        ``make(x=b'\x05')`` is correct.

        """
        pattr = bytes([0xFF])
        for k,v in kwargs.items():
            if not isinstance(v, bytes):
                raise TypeError('Values must be bytes')

            pattr += int_to_bytes(len(k),3) + k.encode()
            pattr += int_to_bytes(len(v),3) + v
        return pattr

    @staticmethod
    def unpack(pattr: bytes) -> Dict[str, bytes]:
        """
        Will parse PackedAttributes.pack
        bytestring and convert it to the
        python dictionary.

        Every PackedAttributes bytestring
        must contain ``0xFF`` as first byte.
        If not, or if error, will return ``{}``.
        """
        if not pattr:
            return {}
        try:
            assert pattr[0] == 0xFF
            pattr_d, pattr = {}, pattr[1:]

            while pattr:
                key_len = bytes_to_int(pattr[:3])
                key = pattr[3:key_len+3]

                value_len = bytes_to_int(pattr[key_len+3:key_len+6])
                pattr = pattr[key_len+6:]

                if value_len:
                    value = pattr[:value_len]
                    pattr = pattr[value_len:]
                else:
                    value = b''

                pattr_d[key.decode()] = value

            return pattr_d
        except:
            return {}

class OpenPretender:
    """
    Class to wrap Tgbox AES Generators and make it look
    like opened to "rb"-read file.
    """
    def __init__(
            self, flo: BinaryIO,
            aes_state: AES,
            file_size: Optional[int] = None
        ):
        """
        Arguments:
            flo (``BinaryIO``):
                File-like object. Like ``open('file','rb')``.

            aes_state (``AESwState``):
                ``AESwState`` with Key and IV.
        """
        self._aes_state = aes_state
        self._flo = flo

        self._buffered_bytes = b''
        self._total_size = file_size
        self._stop_iteration = False

        self._position = 0

    def concat_metadata(self, metadata: bytes) -> None:
        """Concates metadata to the file as (metadata + file)."""
        if self._position:
            raise ConcatError('Concat must be before any usage of object.')
        else:
            self._buffered_bytes += metadata

    async def read(self, size: int=-1) -> bytes:
        """
        Returns ``size`` bytes from async Generator.

        This method is async only because we use
        ``File`` uploading from the async library. You
        can use ``tgbox.sync`` in your sync code for reading.

        Arguments:
            size (``int``):
                Amount of bytes to return. By
                default is negative (return all).
        """
        if self._stop_iteration:
            raise Exception('Stream was closed')

        if size % 16 and not size == -1:
            raise ValueError('size must be divisible by 16 or -1 (return all)')

        if self._total_size is None:
            self._total_size = self._flo.seek(0,2) # Move to file end
            self._flo.seek(0,0) # Move to file start

        if self._total_size <= 0 or size <= len(self._buffered_bytes) and size != -1:
            block = self._buffered_bytes[:size]
            self._buffered_bytes = self._buffered_bytes[size:]
        else:
            buffered = self._buffered_bytes
            self._buffered_bytes = b''

            if size == -1:
                chunk = self._flo.read()
                chunk = await chunk if iscoroutine(chunk) else chunk

                block = buffered + self._aes_state.encrypt(
                    chunk, pad=True, concat_iv=False)
            else:
                chunk = self._flo.read(size)
                chunk = await chunk if iscoroutine(chunk) else chunk

                if len(chunk) % 16:
                    shift = int(-(len(chunk) % 16))
                else:
                    shift = None

                if self._total_size <= 0 or size > self._total_size or shift != None:
                    chunk = buffered + self._aes_state.encrypt(
                        chunk, pad=True, concat_iv=False)
                else:
                    chunk = buffered + self._aes_state.encrypt(
                        chunk, pad=False, concat_iv=False)

                shift = size if len(chunk) > size else None

                if shift is not None:
                    self._buffered_bytes = chunk[shift:]

                self._total_size -= size
                block = chunk[:shift]

        self._position += len(block)
        return block

    def tell(self) -> int:
        return self._position

    def seekable(self, *args, **kwargs) -> bool:
        return False

    def close(self) -> None:
        self._stop_iteration = True

def pad_request_size(request_size: int, bsize: int=4096) -> int:
    """
    This function pads ``request_size`` to divisible
    by BSIZE bytes. If ``request_size`` < BSIZE, then
    it's not padded. This function designed for Telethon's
    ``GetFileRequest``. See issue #3 on TGBOX GitHub.

    .. note::
        You will need to strip extra bytes from result, as
        request_size can be bigger than bytecount you want.

    Arguments:
        request_size (``int``):
            Amount of requested bytes,
            max is 1048576 (1MiB).

        bsize (``int``, optional):
            Size of block. Typically we
            don't need to change this.
    """
    assert request_size <= 1048576, 'Max 1MiB'

    if request_size <= bsize:
        return request_size

    while 1048576 % request_size:
        request_size = ((request_size + bsize) // bsize) * bsize
    return request_size

def ppart_id_generator(path: Path, mainkey: MainKey) -> Generator[tuple, None, None]:
    """
    This generator will iterate over path parts and
    yield their unique IDs. We will use this to better
    navigate over *abstract* Folders in the LocalBox.

    The path **shouldn't** contain a file name,
    otherwise directory will contain it as folder.

    */home/user/* is **OK**
    */home/user/file.txt* is **NOT**

    Will yield a tuple (PART, PARENT_PART_ID, PART_ID)
    """
    parent_part_id = b'' # The root (/ anchor) doesn't have parent
    for part in path.parts:
        part_id = sha256(
            mainkey\
          + sha256(part.encode()).digest()\
          + parent_part_id
        )
        yield (part, parent_part_id, part_id.digest())
        parent_part_id = part_id.digest()

def prbg(size: int) -> bytes:
    """Will generate ``size`` pseudo-random bytes."""
    return bytes([randrange(256) for _ in range(size)])

def int_to_bytes(
        int_: int, length: Optional[int] = None,
        signed: Optional[bool] = False) -> bytes:

    """Converts int to bytes with Big byteorder."""

    bit_length = int_.bit_length()

    if not length:
        if signed and not (int_ >= -128 and int_ <= 127):
            divide_with = 16
        else:
            divide_with = 8

        bit_length = ((bit_length + divide_with) // divide_with)
        length = (bit_length * divide_with) // 8

    return int.to_bytes(int_, length, 'big', signed=signed)

def bytes_to_int(bytes_: bytes, signed: Optional[bool] = False) -> int:
    """Converts bytes to int with Big byteorder."""
    return int.from_bytes(bytes_, 'big', signed=signed)

async def get_media_duration(file_path: str) -> int:
    """Returns video/audio duration with ffmpeg in seconds."""
    func = partial(subprocess_run,
        args=[FFMPEG, '-i', file_path],
        stdout=None, stderr=PIPE
    )
    try:
        loop = get_event_loop()
        stderr = (await loop.run_in_executor(None, func)).stderr
        duration = re_search(b'Duration: (.)+,', stderr).group()
        d = duration.decode().split('.')[0].split(': ')[1].split(':')
        return int(d[0]) * 60**2 + int(d[1]) * 60 + int(d[2])
    except Exception as e:
        raise DurationImpossible(f'Can\'t get media duration: {e}') from None

async def make_media_preview(
        file_path: PathLike,
        temp_path: Optional[PathLike] = None,
        x: int=128, y: int=-1) -> BinaryIO:
    """
    Makes x:y sized thumbnail of the
    video/audio with ffmpeg. "-1"
    preserves one of side size.
    """
    temp_path = Path() if not temp_path else temp_path
    thumbnail_path = Path(temp_path, prbg(4).hex()+'.jpg')

    func = partial(subprocess_run,
        args=[
            FFMPEG, '-i', file_path, '-filter:v', f'scale={x}:{y}', '-an',
            '-loglevel', 'quiet', '-q:v', '2', thumbnail_path
        ],
        stdout=PIPE,
        stderr=None
    )
    try:
        loop = get_event_loop()
        await loop.run_in_executor(None, func)
        thumb = BytesIO(open(thumbnail_path,'rb').read())
        remove_file(thumbnail_path); return thumb
    except Exception as e:
        # If something goes wrong then file is not created (FileNotFoundError)
        raise PreviewImpossible(f'Can\'t make thumbnail: {e}') from None



"""
Copied from Mautrix-Telegram
    github.com/tulir/mautrix-telegram/

Copied again from Painor GitHub
    gist.github.com/painor/7e74de80ae0c819d3e9abcf9989a8dd6

| This file was patched for TGBOX project (github.com/NonProjects/tgbox)
| and may not work as you expected in your libraries.

Big thanks to all contributors of this module.
"""

import os
import math
import asyncio
import hashlib
import inspect
import logging

from typing import (
    Optional, List, AsyncGenerator,
    Union, Awaitable, BinaryIO,
    DefaultDict, Tuple
)
from collections import defaultdict

from telethon.tl.functions.auth import (
    ExportAuthorizationRequest,
    ImportAuthorizationRequest
)
from telethon.errors import FilePartsInvalidError
from telethon import utils, helpers, TelegramClient
from telethon.tl.custom.file import File

from telethon.crypto import AuthKey
from telethon.network import MTProtoSender

from telethon.tl.alltlobjects import LAYER
from telethon.tl.functions import InvokeWithLayerRequest

from telethon.tl.types import (
    Document, InputFileLocation,
    InputDocumentFileLocation,
    InputPhotoFileLocation,
    InputPeerPhotoFileLocation,
    TypeInputFile, Photo,
    InputFileBig, InputFile
)
from telethon.tl.functions.upload import (
    GetFileRequest,
    SaveFilePartRequest,
    SaveBigFilePartRequest
)
log: logging.Logger = logging.getLogger("telethon")

TypeLocation = Union[
    Document, Photo, InputDocumentFileLocation,
    InputPeerPhotoFileLocation,
    InputFileLocation, InputPhotoFileLocation
]
class DownloadSender:
    client: TelegramClient
    sender: MTProtoSender
    request: GetFileRequest
    remaining: int
    stride: int

    def __init__(
            self, client: TelegramClient,
            sender: MTProtoSender,
            file: TypeLocation,
            offset: int, limit: int,
            stride: int, count: int) -> None:

        # TODO: Add left bytes to buffer and return.
        self.offset = offset
        # Offset must be divisible by 4096, otherwise
        # error. We concat left bytes later.
        self.safe_offset = offset - offset % 4096

        self.sender = sender
        self.client = client
        self.stride = stride
        self.remaining = count

        self.request = GetFileRequest(
            file, offset=self.safe_offset, limit=limit
        )

    async def next(self) -> Optional[bytes]:
        if not self.remaining:
            return None

        result = await self.client._call(self.sender, self.request)

        self.remaining -= 1
        self.request.offset += self.stride

        result_bytes = result.bytes[self.offset - self.safe_offset:]
        return result_bytes

    def disconnect(self) -> Awaitable[None]:
        return self.sender.disconnect()


class UploadSender:
    client: TelegramClient
    sender: MTProtoSender
    request: Union[SaveFilePartRequest, SaveBigFilePartRequest]
    part_count: int
    stride: int
    previous: Optional[asyncio.Task]
    loop: asyncio.AbstractEventLoop

    def __init__(self, client: TelegramClient, sender: MTProtoSender, file_id: int, part_count: int, big: bool,
                 index: int,
                 stride: int, loop: asyncio.AbstractEventLoop) -> None:
        self.client = client
        self.sender = sender
        self.part_count = part_count
        if big:
            self.request = SaveBigFilePartRequest(file_id, index, part_count, b"")
        else:
            self.request = SaveFilePartRequest(file_id, index, b"")
        self.stride = stride
        self.previous = None
        self.loop = loop

    async def next(self, data: bytes) -> None:
        if self.previous:
            await self.previous
        self.previous = self.loop.create_task(self._next(data))

    async def _next(self, data: bytes) -> None:
        self.request.bytes = data
        log.debug(f"Sending file part {self.request.file_part}/{self.part_count}"
                  f" with {len(data)} bytes")
        await self.client._call(self.sender, self.request)
        self.request.file_part += self.stride

    async def disconnect(self) -> None:
        if self.previous:
            await self.previous
        return await self.sender.disconnect()


class ParallelTransferrer:
    client: TelegramClient
    loop: asyncio.AbstractEventLoop
    dc_id: int
    senders: Optional[List[Union[DownloadSender, UploadSender]]]
    auth_key: AuthKey
    upload_ticker: int

    def __init__(self, client: TelegramClient, dc_id: Optional[int] = None) -> None:
        self.client = client
        self.loop = self.client.loop
        self.dc_id = dc_id or self.client.session.dc_id
        self.auth_key = (None if dc_id and self.client.session.dc_id != dc_id
                         else self.client.session.auth_key)
        self.senders = None
        self.upload_ticker = 0

    async def _cleanup(self) -> None:
        try:
            await asyncio.gather(*[sender.disconnect() for sender in self.senders])
        except FilePartsInvalidError:
            pass
        self.senders = None

    @staticmethod
    def _get_connection_count(file_size: int, max_count: int = 20,
                              full_size: int = 100 * 1024 * 1024) -> int:
        if file_size > full_size:
            return max_count
        return math.ceil((file_size / full_size) * max_count)

    async def _init_download(
            self, connections: int,
            file: TypeLocation,
            part_count: int,
            part_size: int,
            offset: int) -> None:

        minimum, remainder = divmod(part_count, connections)

        def get_part_count() -> int:
            nonlocal remainder
            if remainder > 0:
                remainder -= 1
                return minimum + 1
            return minimum

        # The first cross-DC sender will export+import the authorization, so
        # we always create it before creating any other senders.
        self.senders = [
            await self._create_download_sender(
                file, 0, part_size, connections * part_size,
                get_part_count(), offset=offset
            ),
            *await asyncio.gather(
                *[
                    self._create_download_sender(
                        file, i, part_size,
                        connections * part_size,
                        get_part_count(),
                        offset=offset
                    )
                    for i in range(1, connections)
                ])
        ]
    async def _create_download_sender(
            self, file: TypeLocation,
            index: int,
            part_size: int,
            stride: int,
            part_count: int,
            offset: int = None) -> DownloadSender:

        offset = offset if offset else index * part_size

        return DownloadSender(
            self.client,
            await self._create_sender(),
            file, offset, part_size,
            stride, part_count
        )

    async def _init_upload(self, connections: int, file_id: int, part_count: int, big: bool
                           ) -> None:
        self.senders = [
            await self._create_upload_sender(file_id, part_count, big, 0, connections),
            *await asyncio.gather(
                *[self._create_upload_sender(file_id, part_count, big, i, connections)
                  for i in range(1, connections)])
        ]

    async def _create_upload_sender(self, file_id: int, part_count: int, big: bool, index: int,
                                    stride: int) -> UploadSender:
        return UploadSender(self.client, await self._create_sender(), file_id, part_count, big, index, stride,
                            loop=self.loop)

    async def _create_sender(self) -> MTProtoSender:
        dc = await self.client._get_dc(self.dc_id)
        sender = MTProtoSender(self.auth_key, loggers=self.client._log)
        await sender.connect(self.client._connection(dc.ip_address, dc.port, dc.id,
                                                     loggers=self.client._log,
                                                     proxy=self.client._proxy))
        if not self.auth_key:
            log.debug(f"Exporting auth to DC {self.dc_id}")
            auth = await self.client(ExportAuthorizationRequest(self.dc_id))
            self.client._init_request.query = ImportAuthorizationRequest(id=auth.id,
                                                                         bytes=auth.bytes)
            req = InvokeWithLayerRequest(LAYER, self.client._init_request)
            await sender.send(req)
            self.auth_key = sender.auth_key
        return sender

    async def init_upload(self, file_id: int, file_size: int, part_size_kb: Optional[float] = None,
                          connection_count: Optional[int] = None) -> Tuple[int, int, bool]:
        connection_count = connection_count or self._get_connection_count(file_size)
        part_size = (part_size_kb or utils.get_appropriated_part_size(file_size)) * 1024
        part_count = (file_size + part_size - 1) // part_size
        is_large = file_size > 10 * 1024 * 1024
        await self._init_upload(connection_count, file_id, part_count, is_large)
        return part_size, part_count, is_large

    async def upload(self, part: bytes) -> None:
        await self.senders[self.upload_ticker].next(part)
        self.upload_ticker = (self.upload_ticker + 1) % len(self.senders)

    async def finish_upload(self) -> None:
        await self._cleanup()

    async def download(
            self, file: TypeLocation, file_size: int,
            part_size_kb: Optional[int] = None,
            offset: Optional[int] = None,
            connection_count: Optional[int] = None
            ) -> AsyncGenerator[bytes, None]:

        connection_count = connection_count or self._get_connection_count(file_size)
        part_size = (part_size_kb or utils.get_appropriated_part_size(file_size)) * 1024
        part_count = math.ceil(file_size / part_size)

        log.debug("Starting parallel download: "
                  f"{connection_count} {part_size} {part_count} {file!s}")

        await self._init_download(
            connection_count,
            file, part_count,
            part_size, offset=offset
        )
        part = 0
        while part < part_count:
            tasks = []
            for sender in self.senders:
                tasks.append(self.loop.create_task(sender.next()))
            for task in tasks:
                data = await task
                if not data:
                    break
                yield data
                part += 1
                log.debug(f"Part {part} downloaded")

        log.debug("Parallel download finished, cleaning up connections")
        await self._cleanup()


parallel_transfer_locks: DefaultDict[int, asyncio.Lock] =\
    defaultdict(lambda: asyncio.Lock())

async def stream_file(
        file_to_stream: BinaryIO,
        chunk_size = 1024
    ):
    """``file_to_stream.read`` can be coroutine."""
    while True:
        data_read = file_to_stream.read(chunk_size)

        if inspect.iscoroutine(data_read):
            data_read = await data_read

        if data_read:
            yield data_read
        else:
            break

async def _internal_transfer_to_telegram(
        client: TelegramClient,
        response: BinaryIO,
        progress_callback: callable,
        file_size: int = None,
        file_name: str = 'document',
        part_size_kb: int = 512,
    ) -> Tuple[TypeInputFile, int]:

    file_id = helpers.generate_random_long()

    if not file_size:
        file_size = os.path.getsize(response.name)

    hash_md5 = hashlib.md5()
    uploader = ParallelTransferrer(client)

    part_size_, part_count, is_large =\
        await uploader.init_upload(
            file_id, file_size,
            part_size_kb=part_size_kb
        )
    if not part_size_kb:
        part_size = part_size_
    else:
        part_size = part_size_kb*1024

    buffer = bytearray()
    async for data in stream_file(response, part_size):
        if progress_callback:
            r = progress_callback(response.tell(), file_size)
            if inspect.isawaitable(r):
                await r

        if not is_large:
            hash_md5.update(data)

        if len(buffer) == 0 and len(data) == part_size:
            try:
                await uploader.upload(data)
                continue
            except FilePartsInvalidError as e:
                await uploader.finish_upload()
                raise e from None

        new_len = len(buffer) + len(data)
        if new_len >= part_size:
            cutoff = part_size - len(buffer)
            buffer.extend(data[:cutoff])

            await uploader.upload(bytes(buffer))

            buffer.clear()
            buffer.extend(data[cutoff:])
        else:
            buffer.extend(data)

    if len(buffer) > 0:
        await uploader.upload(bytes(buffer))

    await uploader.finish_upload()

    if is_large:
        return (InputFileBig(file_id, part_count, file_name), file_size)
    else:
        return (InputFile(file_id, part_count, file_name, hash_md5.hexdigest()), file_size)

async def download_file(
        client: TelegramClient,
        location: TypeLocation,
        request_size: int=524288,
        offset: int=None,
        progress_callback: callable = None
        ) -> AsyncGenerator[bytes, None]:

    if isinstance(location, Photo):
        size = File(location).size
    else:
        size = location.size

    dc_id, location = utils.get_input_location(location)

    # We lock the transfers because telegram has connection count limits
    downloader = ParallelTransferrer(client, dc_id)

    downloaded = downloader.download(
        location, size, offset=offset,
        part_size_kb=int(request_size/1024)
    )
    position = 0
    async for chunk in downloaded:
        position += len(chunk)
        if progress_callback:
            r = progress_callback(position, size)
            if inspect.isawaitable(r):
                await r
        yield chunk


async def upload_file(
        client: TelegramClient,
        file: BinaryIO,
        progress_callback: callable = None,
        file_name: str = 'document',
        file_size: int = None,
        part_size_kb: int = 512
        ) -> TypeInputFile:

    return (await _internal_transfer_to_telegram(
        client, file, progress_callback,
        file_size = file_size,
        file_name = file_name,
        part_size_kb = part_size_kb
        ))[0]



"""Module with API functions and classes for LocalBox."""

from typing import (
    BinaryIO, Union, NoReturn, Callable,
    AsyncGenerator, Dict, Optional
)
from os.path import getsize
from pathlib import Path

from os import PathLike
from io import BytesIO
from time import time

from hashlib import sha256
from asyncio import iscoroutinefunction, gather

from filetype import guess as filetype_guess
from telethon.tl.types import Photo, Document

from ..crypto import get_rnd_bytes
from ..crypto import AESwState as AES

from ..keys import (
    make_filekey, make_requestkey,
    EncryptedMainkey, make_mainkey,
    make_sharekey, MainKey, RequestKey,
    ShareKey, ImportKey, FileKey, BaseKey
)
from ..defaults import PREFIX, VERBYTE, DEF_TGBOX_NAME

from ..errors import (
    NotInitializedError, AlreadyImported,
    InUseException, AESError, PreviewImpossible,
    LimitExceeded, DurationImpossible, InvalidFile,
    NotEnoughRights, IncorrectKey, FingerprintExists
)
from ..tools import (
    PackedAttributes, ppart_id_generator,
    int_to_bytes, bytes_to_int, SearchFilter,
    get_media_duration, prbg, make_media_preview
)
from .utils import (
    DirectoryRoot, search_generator, PreparedFile,
    _TelegramVirtualFile, TelegramClient,
    DefaultsTableWrapper, RemoteBoxDefaults
)
from .db import TgboxDB

__all__ = [
    'make_localbox',
    'get_localbox',
    'clone_remotebox',
    'EncryptedLocalBox',
    'DecryptedLocalBox',
    'EncryptedLocalBoxDirectory',
    'DecryptedLocalBoxDirectory',
    'EncryptedLocalBoxFile',
    'DecryptedLocalBoxFile',
]
async def make_localbox(
        erb: 'tgbox.api.remote.EncryptedRemoteBox',
        basekey: BaseKey) -> 'DecryptedLocalBox':
    """
    Makes LocalBox

    Arguments:
        erb (``RemoteBox``):
            ``EncryptedRemoteBox``. You will
            recieve it after ``make_remotebox``.

        basekey (``BaseKey``):
            ``BaseKey`` that will be used
            for ``MainKey`` creation.
    """
    tgbox_db = await TgboxDB.create(await erb.get_box_name())

    if (await tgbox_db.BOX_DATA.count_rows()):
        raise InUseException(
           f'''"{tgbox_db.name}" was found on current path. '''
            '''Please move old file or create RemoteBox with '''
            '''the different box name (see help(tgbox.api.re'''
            '''mote.make_remotebox) and use kwarg "box_name")'''
        )
    box_salt = await erb.get_box_salt()
    mainkey = make_mainkey(basekey, box_salt)

    await tgbox_db.BOX_DATA.insert(
        AES(mainkey).encrypt(int_to_bytes(erb._box_channel_id)),
        AES(mainkey).encrypt(int_to_bytes(int(time()))),
        box_salt,
        None, # We aren't cloned box, so Mainkey is empty
        AES(basekey).encrypt(erb._tc.session.save().encode()),
        AES(mainkey).encrypt(int_to_bytes(erb._tc._api_id)),
        AES(mainkey).encrypt(bytes.fromhex(erb._tc._api_hash)),
    )
    return await EncryptedLocalBox(tgbox_db).decrypt(basekey)

async def get_localbox(
        basekey: Optional[BaseKey] = None,
        tgbox_db_path: Optional[Union[PathLike, str]] = DEF_TGBOX_NAME,
        ) -> Union['EncryptedLocalBox', 'DecryptedLocalBox']:
    """
    Returns LocalBox.

    Arguments:
        basekey (``BaseKey``, optional):
            Returns ``DecryptedLocalBox`` if specified,
            ``EncryptedLocalBox`` otherwise (default).

        tgbox_db_path (``PathLike``, ``str``, optional):
            ``PathLike`` to your TgboxDB (LocalBox). Default
            is ``defaults.DEF_TGBOX_NAME``.
    """
    if not isinstance(tgbox_db_path, PathLike):
        tgbox_db_path = Path(tgbox_db_path)

    if not tgbox_db_path.exists():
        raise FileNotFoundError(f'Can\'t open {tgbox_db_path.absolute()}')
    else:
        tgbox_db = await TgboxDB(tgbox_db_path).init()

    if basekey:
        return await EncryptedLocalBox(tgbox_db).decrypt(basekey)
    else:
        return await EncryptedLocalBox(tgbox_db).init()

async def clone_remotebox(
        drb: 'tgbox.api.remote.DecryptedRemoteBox',
        basekey: BaseKey,
        progress_callback: Optional[Callable[[int, int], None]] = None,
        box_path: Optional[Union[PathLike, str]] = None) -> 'DecryptedLocalBox':
    """
    This method makes ``LocalBox`` from ``RemoteBox`` and
    imports all RemoteBoxFiles to it.

    Arguments:
        drb (``DecryptedRemoteBox``):
            ``DecryptedRemoteBox`` you want to clone.

        basekey (``BaseKey``):
            ``BaseKey`` with which you will decrypt your
            cloned ``EncryptedLocalBox``. ``BaseKey`` encrypts
            Session and ``MainKey`` of original LocalBox.

        progress_callback (``Callable[[int, int], None]``, optional):
            A callback function accepting two parameters:
            (current_amount, total).

        box_path (``PathLike``, ``str``, optional):
            Direct path with filename included. If
            not specified, then ``RemoteBox`` name used.
    """
    box_path = await drb.get_box_name()\
        if not box_path else box_path

    tgbox_db = await TgboxDB.create(box_path)

    if (await tgbox_db.BOX_DATA.count_rows()):
        raise InUseException(f'TgboxDB "{tgbox_db.name}" in use. Specify new.')

    last_file_id = 0
    async for erbf in drb.files(decrypt=False, return_imported_as_erbf=True):
        last_file_id = erbf.id; break

    await tgbox_db.BOX_DATA.insert(
        AES(drb._mainkey).encrypt(int_to_bytes(drb._box_channel_id)),
        AES(drb._mainkey).encrypt(int_to_bytes(int(time()))),
        await drb.get_box_salt(),
        AES(basekey).encrypt(drb._mainkey.key),
        AES(basekey).encrypt(drb._tc.session.save().encode()),
        AES(drb._mainkey).encrypt(int_to_bytes(drb._tc._api_id)),
        AES(drb._mainkey).encrypt(bytes.fromhex(drb._tc._api_hash)),
    )
    dlb = await EncryptedLocalBox(tgbox_db).decrypt(basekey)

    files_generator = drb.files(
        key=drb._mainkey,
        decrypt=True, reverse=True,
        erase_encrypted_metadata=False
    )
    async for drbf in files_generator:
        if progress_callback:
            if iscoroutinefunction(progress_callback):
                await progress_callback(drbf.id, last_file_id)
            else:
                progress_callback(drbf.id, last_file_id)

        await dlb.import_file(drbf)

    return dlb

class EncryptedLocalBox:
    """
    This class represents an encrypted local box. On more
    low-level that's a wrapper around ``TgboxDB``. Usually
    you will never meet this class in your typical code,
    but you may want to extract some encrypted data.

    Usage:

    .. code-block:: python

        from asyncio import run as asyncio_run
        from tgbox.api import EncryptedLocalBox
        from tgbox.db import TgboxDB

        async def main():
            # Make or open TgboxDB
            tdb = await TgboxDB.create('TGBOX')
            # Initialize EncryptedLocalBox
            elb = await EncryptedLocalBox(tdb).init()

            # Retrieve encrypted session or None
            print(elb.session)

        asyncio_run(main())

    You can acces it from the ``DecryptedLocalBox``:

    .. code-block:: python

        from asyncio import run as asyncio_run
        from tgbox.api import get_localbox
        from tgbox.keys import make_basekey, Phrase

        phrase = Phrase(b'example phrase here')
        basekey = make_basekey(phrase)

        async def main():
            dlb = await get_localbox(basekey)
            # Retrieve encrypted session
            print(dlb._elb.session)

        asyncio_run(main())
    """
    def __init__(
            self, tgbox_db: TgboxDB,
            defaults: Optional[Union[DefaultsTableWrapper,
                RemoteBoxDefaults]] = None):
        """
        Arguments:
            tgbox_db (``TgboxDB``):
                Initialized Tgbox Database.

            defaults (``DefaultsTableWrapper``, ``RemoteBoxDefaults``):
                Class with a default values/constants we will use.
        """
        self._tgbox_db = tgbox_db

        if defaults is None:
            self._defaults = DefaultsTableWrapper(self._tgbox_db)
        else:
            self._defaults = defaults

        self._api_id = None
        self._api_hash = None

        self._mainkey = None
        self._box_salt = None
        self._session = None

        self._box_channel_id = None
        self._box_cr_time = None

        self._initialized = False

    def __hash__(self) -> int:
        if not self._initialized:
            raise NotInitializedError(
                'Must be initialized before hashing'
            )
        # Session will be different in Enc or Dec classes.
        return hash((self._box_salt, self._session))

    def __eq__(self, other) -> bool:
        return all((
            isinstance(other, self.__class__),
            self.__hash__() == hash(other)
        ))
    def __raise_initialized(self) -> NoReturn:
        if not self._initialized:
            raise NotInitializedError('Not initialized. Call .init().')

    @property
    def api_id(self) -> Union[int, None]:
        """Returns API_ID."""
        return self._api_id

    @property
    def api_hash(self) -> Union[str, None]:
        """Returns API_HASH."""
        return self._api_hash

    @property
    def tgbox_db(self) -> TgboxDB:
        """Returns ``TgboxDB``."""
        return self._tgbox_db

    @property
    def defaults(self) -> Union[DefaultsTableWrapper, RemoteBoxDefaults]:
        """Returns ``DefaultsTableWrapper`` or ``RemoteBoxDefaults``."""
        return self._defaults

    @property
    def initialized(self) -> bool:
        """Returns ``True`` if you called ``.init()``"""
        return self._initialized

    @property
    def box_salt(self) -> Union[bytes, None]:
        """Returns BoxSalt or ``None`` if not initialized"""
        return self._box_salt

    @property
    def session(self) -> Union[bytes, str, None]:
        """
        Returns encrypted session from
        ``EncryptedLocalBox`` and decrypted
        from ``DecryptedLocalBox``.
        """
        return self._session

    @property
    def box_channel_id(self) -> Union[bytes, int, None]:
        """
        Returns encrypted channel ID from
        ``EncryptedLocalBox`` and decrypted
        from ``DecryptedLocalBox``.
        """
        return self._box_channel_id

    @property
    def box_cr_time(self) -> Union[bytes, int, None]:
        """
        Returns encrypted box creation time from
        ``EncryptedLocalBox`` and decrypted
        from ``DecryptedLocalBox``.
        """
        return self._box_cr_time

    async def get_last_file_id(self) -> int:
        """
        Returns last file id from the FILES table.
        If there is no files at all, will return 0.
        """
        try:
            lfi = await self._tgbox_db.FILES.select_once((
                'SELECT ID FROM FILES ORDER BY ID DESC LIMIT 1', ()
            ))
            return lfi[0]
        except StopAsyncIteration:
            return 0

    async def init(self) -> 'EncryptedLocalBox':
        """Will fetch and parse data from Database."""

        if not await self._tgbox_db.BOX_DATA.count_rows():
            raise NotInitializedError('Table is empty.')
        else:
            box_data = await self._tgbox_db.BOX_DATA.select_once()
            self._box_channel_id = box_data[0]
            self._box_cr_time, self._box_salt, self._mainkey = box_data[1:4]
            self._session, self._initialized = box_data[4], True
            self._api_id, self._api_hash = box_data[5], box_data[6]

            if self._mainkey:
                self._mainkey = EncryptedMainkey(self._mainkey)

            if isinstance(self._defaults, DefaultsTableWrapper):
                if not self._defaults.initialized:
                    await self._defaults.init()

        return self

    async def get_file(
            self, id: int, decrypt: bool=True, cache_preview: bool=True)\
            -> Union['DecryptedLocalBoxFile', 'EncryptedLocalBoxFile', None]:
        """
        Returns ``EncryptedLocalBoxFile`` from ``EncryptedLocalBox``
        or ``DecryptedLocalBoxFile`` from ``DecryptedLocalBox`` if
        file exists. ``None`` otherwise.

        Arguments:
            id (``int``):
                File ID.

            decrypt (``bool``, optional):
                Force return EncryptedLocalBoxFile

            cache_preview (``bool``, optional):
                Cache preview in class or not.
        """
        try:
            self.__raise_initialized()
            elbf = EncryptedLocalBoxFile(
                id, self._tgbox_db,
                cache_preview=cache_preview,
                defaults=self._defaults
            )
            if decrypt and self._mainkey and not\
                isinstance(self._mainkey, EncryptedMainkey):
                    return await elbf.decrypt(self._mainkey)
            else:
                return await elbf.init()
        except StopAsyncIteration: # No file by ``id``.
            return None

    async def contents(
            self, sfpid: Optional[bytes] = None,
            ignore_files: Optional[bool] = False
                ) -> AsyncGenerator[Union[
                'EncryptedLocalBoxDirectory',
                'DecryptedLocalBoxDirectory'], None
            ]:
        """
        Recursive iterate over all files/folders in LocalBox.

        Arguments:
            sfpid (``bytes``, optional):
                Will start from this PartID if specified,
                else will start from "root" PIDs.

            ignore_files (``bool``, optional):
                Will **not** return LocalBoxFile associated
                with the *LocalBoxDirectory* if ``False``.
        """
        sfpid = (sfpid,) if sfpid else []

        if not sfpid:
            root_pids = await self._tgbox_db.PATH_PARTS.execute((
                'SELECT PART_ID FROM PATH_PARTS WHERE PARENT_PART_ID IS NULL', ()
            ))
            sfpid = [i[0] for i in await root_pids.fetchall()]

        for pid in sfpid:
            lbfid = EncryptedLocalBoxDirectory(self._tgbox_db, pid)

            if isinstance(self, DecryptedLocalBox):
                lbfid = await lbfid.decrypt(self._mainkey)
            else:
                lbfid = await lbfid.init()

            yield lbfid

            if not ignore_files:
                async for lbfi in lbfid.iterdir(ignore_dirs=True):
                    yield lbfi

            child_pids = await self._tgbox_db.PATH_PARTS.execute((
                'SELECT PART_ID FROM PATH_PARTS WHERE PARENT_PART_ID IS ?',
                (pid,)
            ))
            child_sfpid = [i[0] for i in await child_pids.fetchall()]

            for csfpid in child_sfpid:
                contents = self.contents(csfpid, ignore_files=ignore_files)
                async for content in contents:
                    yield content

    async def files(
            self, cache_preview: bool=True,
            min_id: Optional[int] = None,
            max_id: Optional[int] = None)\
            -> Union[
                'DecryptedLocalBoxFile',
                'EncryptedLocalBoxFile', None
               ]:
        """
        Yields every local file as ``EncryptedLocalBoxFile`` if you
        call it on ``EncryptedLocalBox`` and ``DecryptedLocalBoxFile``
        if on ``DecryptedLocalBox``. Works via ``self.get_file``

        Alternatively, you may use LocalBox.contents method.

        Arguments:
            cache_preview (``bool``, optional):
                Cache preview in class or not.

            min_id (``bool``, optional):
                Will iterate from this ID.

            max_id (``bool``, optional):
                Will iterate up to this ID.
        """
        min_id = f'ID > {min_id}' if min_id else ''
        max_id = f'ID < {max_id}' if max_id else ''

        min_id = min_id + ' AND' if all((min_id,max_id)) else min_id
        where = 'WHERE' if any((min_id, max_id)) else ''

        sql_query = f'SELECT ID FROM FILES {where} {min_id} {max_id}'
        cursor = await self._tgbox_db.FILES.execute((sql_query ,()))

        while True:
            pending = await cursor.fetchmany(100)
            if not pending: return # No more files

            pending = [
                self.get_file(file_id[0], cache_preview=cache_preview)
                for file_id in pending
            ]
            pending = await gather(*pending)

            while pending:
                yield pending.pop(0)

    def get_requestkey(self, basekey: BaseKey) -> RequestKey:
        """
        Returns ``RequestKey`` for this *LocalBox*.
        You should use this method if you want
        to decrypt other's ``RemoteBox``.

        Arguments:
            basekey (``BaseKey``):
                To make a ``RequestKey`` for other's ``RemoteBox``
                you need to create new ``BaseKey`` for it. Later
                this key will be used for *Box* decryption.
        """
        self.__raise_initialized()
        return make_requestkey(basekey, box_salt=self._box_salt)

    async def delete(self) -> None:
        """
        This method **WILL DELETE** your *LocalBox*
        database. It doesn't affect *RemoteBox*,
        so you can make new *LocalBox* from the
        *Remote* version if you have *MainKey*.

        Will raise ``FileNotFoundError`` if
        something goes wrong (i.e DB was moved).
        """
        if not self._tgbox_db.closed:
            await self.done()

        self._tgbox_db.db_path.unlink()

    async def decrypt(self, key: Union[BaseKey, MainKey]) -> 'DecryptedLocalBox':
        """Will return ``DecryptedLocalBox``.

        You **should** specify ``BaseKey`` if you
        want to access session and use this dlb
        with *RemoteBox*, however, you can specify
        ``MainKey`` if you only want to iterate
        over local files / fetch basic local info.
        """
        if not self.initialized: await self.init()
        return DecryptedLocalBox(self, key)

    async def done(self):
        """
        Await this method when you end all
        work with LocalBox, so we will
        clean up & close connections.
        """
        await self._tgbox_db.close()

class DecryptedLocalBox(EncryptedLocalBox):
    """
    Class that represents decrypted local box. On
    more low-level it's wrapper around ``TgboxDB`` that
    decrypts and parses every row. You don't need to
    work with ``EncryptedLocalBox`` to write any data
    to the ``TgboxDB``. Every commit will be encrypted here.

    Typical usage:

    .. code-block: python

        from asyncio import run as asyncio_run
        from tgbox.api import get_localbox, DecryptedLocalBoxFile
        from tgbox.keys import make_basekey, Phrase

        phrase = Phrase(b'example phrase here')
        basekey = make_basekey(phrase)

        async def main():
            dlb = await get_localbox(basekey)

            # Iterate over LocalBox contents
            async for content in dlb.contents(ignore_files=False):
                if isinstance(content, DecryptedLocalBoxFile):
                    print('File:', file.id, file.file_name, file.size)
                else:
                    await content.lload(full=True) # Load directory
                    print('Dir:', content, content.part_id.hex())

        asyncio_run(main())
    """
    def __init__(self, elb: EncryptedLocalBox, key: Union[BaseKey, MainKey]):
        """
        Arguments:
            elb (``EncryptedLocalBox``):
                Local box you want to decrypt.

            key (``BaseKey``, ``MainKey``):
                You **should** specify ``BaseKey`` if you
                want to access session and use this dlb
                with *RemoteBox*, however, you can specify
                ``MainKey`` if you only want to iterate
                over local files / fetch basic local info.
        """
        if not elb.initialized:
            raise NotInitializedError('Parent class isn\'t initialized.')

        self._elb = elb

        self._tgbox_db = elb._tgbox_db
        self._defaults = elb._defaults

        self._initialized = True

        if isinstance(key, BaseKey):
            if isinstance(elb._mainkey, EncryptedMainkey):
                mainkey = AES(key).decrypt(elb._mainkey.key)
                self._mainkey = MainKey(mainkey)
            else:
                self._mainkey = make_mainkey(key, self._elb._box_salt)
            try:
                # We encrypt Session with Basekey to prevent stealing
                # Session information by people who also have mainkey
                # of the same box. So there is decryption with basekey.
                self._session = AES(key).decrypt(elb._session).decode()
            except UnicodeDecodeError:
                raise IncorrectKey('Can\'t decrypt Session. Invalid Basekey?')

        elif isinstance(key, MainKey):
            self._mainkey = key
            self._session = None
        else:
            raise IncorrectKey('key is not Union[BaseKey, MainKey]')

        self._box_channel_id = bytes_to_int(
            AES(self._mainkey).decrypt(elb._box_channel_id)
        )
        self._box_cr_time = bytes_to_int(
            AES(self._mainkey).decrypt(elb._box_cr_time)
        )
        self._api_id = bytes_to_int(
            AES(self._mainkey).decrypt(elb._api_id)
        )
        self._api_hash = AES(self._mainkey).decrypt(elb._api_hash).hex()
        self._box_salt = elb._box_salt

    @staticmethod
    async def init() -> NoReturn:
        raise AttributeError(
            """This function was inherited from ``EncryptedLocalBox`` """
            """and cannot be used on ``DecryptedLocalBox``."""
        )
    @staticmethod
    async def decrypt() -> NoReturn:
        raise AttributeError(
            """This function was inherited from ``EncryptedLocalBox`` """
            """and cannot be used on ``DecryptedLocalBox``."""
        )
    async def _make_local_path(self, file_path: Path) -> 'DecryptedLocalBoxDirectory':
        """
        Creates abstract LocalBoxDirectory, and returns
        the ``DecryptedLocalBoxDirectory`` object.

        Arguments:
            file_path (``Path``):
                File path. No filename included.
        """
        ppidg = ppart_id_generator(file_path, self._mainkey)

        for part, parent_part_id, part_id in ppidg:
            if not parent_part_id:
                parent_part_id = None

            cursor = await self._tgbox_db.PATH_PARTS.execute((
                'SELECT PART_ID FROM PATH_PARTS WHERE PART_ID=?', (part_id,))
            )
            if not await cursor.fetchone():
                await self._tgbox_db.PATH_PARTS.insert(
                    AES(self._mainkey).encrypt(part.encode()),
                    part_id, parent_part_id
                )
        elbd = EncryptedLocalBoxDirectory(self._tgbox_db, part_id)
        return await elbd.decrypt(self._mainkey)

    async def _make_local_file(self, pf: 'PreparedFile') -> 'DecryptedLocalBoxFile':
        """
        Creates a LocalBoxFile.

        Arguments:
            pf (``PreparedFile``):
                Pushed to RemoteBox ``PreparedFile``.
        """
        assert hasattr(pf,'file_id'), 'Push to RemoteBox firstly'
        assert hasattr(pf,'upload_time'), 'Push to RemoteBox firstly'
        try:
            # Verify that there is no file with the same ID
            await self._tgbox_db.FILES.select_once(
                sql_tuple=('SELECT ID FROM FILES WHERE ID=?', (pf.file_id,))
            )
        except StopAsyncIteration:
            pass
        else:
            raise AlreadyImported('There is already file with same ID') from None

        eupload_time = AES(pf.filekey).encrypt(int_to_bytes(pf.upload_time))

        if pf.imported:
            if make_filekey(self._mainkey, pf.filesalt) == pf.filekey:
                efilekey = None # We can make it with our MainKey
            else:
                efilekey = AES(self._mainkey).encrypt(pf.filekey.key)
        else:
            efilekey = None

        part_id = (await self._make_local_path(pf.filepath)).part_id

        await self._tgbox_db.FILES.insert(
            pf.file_id, eupload_time,
            part_id, efilekey, pf.fingerprint,
            pf.metadata, None
        )
        return await EncryptedLocalBoxFile(
            pf.file_id, self._tgbox_db,
            defaults=self._defaults).decrypt(pf.filekey)

    async def _check_fingerprint(self, fingerprint: bytes):
        """
        Will check that file path is unique, and if not
        then will raise a ``FingerprintExists`` error.

        This is mostly for internal use.

        Arguments:
            fingerprint (``bytes``):
                BoxFile fingerprint.
        """
        try:
            # Verify that there is no file with the same fingerprint
            await self._tgbox_db.FILES.select_once(sql_tuple=(
                'SELECT ID FROM FILES WHERE FINGERPRINT=?',
                (fingerprint,)
            ))
        except StopAsyncIteration:
            pass
        else:
            raise FingerprintExists(FingerprintExists.__doc__) from None

    # TODO: Delete PartID from PATH_PARTS if it links
    # to only one file that was removed.
    async def sync(
            self, drb: 'tgbox.api.remote.DecryptedRemoteBox',
            start_from: int=0,
            progress_callback: Optional[Callable[[int, int], None]] = None):
        """
        This method will synchronize your LocalBox
        with RemoteBox. All files that not in RemoteBox
        but in Local will be **removed**, all that
        in Remote but not in LocalBox will be imported.

        drb (``DecryptedRemoteBox``):
            *RemoteBox* associated with this LocalBox.

        start_from (``int``, optional):
            Will check files that > start_from [ID].

        progress_callback (``Callable[[int, int], None]``, optional):
            A callback function accepting
            two parameters: (last_id, current_id).
        """
        async def _get_file(n=start_from):
            iter_over = drb.files(
                min_id=n, reverse=True,
                cache_preview=False,
                return_imported_as_erbf=True
            )
            async for drbf in iter_over:
                return drbf

        def difference(sql_tuple_ids: tuple) -> bool:
            """
            This local func will sort out useless SQL
            querys, like DELETE FROM FILES WHERE ID > 5 AND ID < 5
            """
            if sql_tuple_ids[0] == sql_tuple_ids[1] \
                or sql_tuple_ids[0]+1 == sql_tuple_ids[1]:
                    return False
            else:
                return True

        async for drbf in drb.files(cache_preview=False):
            last_drbf_id = drbf.id; break

        rbfiles = []

        while True:
            current = 0

            if None in rbfiles:
                break

            if not rbfiles:
                rbfiles.append(await _get_file())

                if rbfiles[0] is None:
                    # RemoteBox doesn't have any files
                    await self._tgbox_db.FILES.execute(
                        sql_tuple=('DELETE FROM FILES', ()))
                    break

                rbfiles.append(await _get_file(rbfiles[0].id))
                last_id = rbfiles[0].id

                sql_tuple = (
                    'DELETE FROM FILES WHERE ID < ?',
                    (last_id,)
                )
                await self._tgbox_db.FILES.execute(
                    sql_tuple=sql_tuple
                )
            else:
                rbfiles.append(await _get_file(rbfiles[1].id))
                if None in rbfiles: break

                rbfiles.append(await _get_file(rbfiles[2].id-1))
                if None in rbfiles: break
                rbfiles.pop(0); rbfiles.pop(1)

            if progress_callback:
                if iscoroutinefunction(progress_callback):
                    await progress_callback(last_id, last_drbf_id)
                else:
                    progress_callback(last_id, last_drbf_id)

            while True:
                if current == 2 or not rbfiles[current]:
                    break
                try:
                    lbfi_id = await self._tgbox_db.FILES.select_once(
                        sql_tuple = (
                            'SELECT ID FROM FILES WHERE ID=?',
                            (rbfiles[current].id,)
                        )
                    )
                except StopAsyncIteration:
                    lbfi_id = None

                if lbfi_id or 'Encrypted' in repr(rbfiles[current]):
                    current += 1
                else:
                    await self.import_file(rbfiles[current])

                    if current == 0:
                        rbfiles[0] = rbfiles[1]

                    if None in rbfiles:
                        break

                    rbfiles[1] = await _get_file(rbfiles[0].id)

                    if rbfiles[0] and not rbfiles[1]:
                        if current == 0:
                            try:
                                await self.import_file(rbfiles[current])
                            except AlreadyImported:
                                pass # Last file may be already in Box, so skip

                        rbfiles[1] = rbfiles[0]
                        break
                    else:
                        last_id = rbfiles[1].id

            sql_tuple = (
                'DELETE FROM FILES WHERE ID > ? AND ID < ?',
                (last_id, rbfiles[0].id)
            )
            if difference(sql_tuple[1]):
                await self._tgbox_db.FILES.execute(sql_tuple=sql_tuple)

            last_id = rbfiles[1].id if rbfiles[1] else None

            if last_id:
                sql_tuple = (
                    'DELETE FROM FILES WHERE ID > ? AND ID < ?',
                    (rbfiles[0].id, rbfiles[1].id)
                )
                if difference(sql_tuple[1]):
                    await self._tgbox_db.FILES.execute(sql_tuple=sql_tuple)
            else:
                sql_tuple = (
                    'DELETE FROM FILES WHERE ID = ?',
                    (rbfiles[0].id,)
                )
                await self._tgbox_db.FILES.execute(sql_tuple=sql_tuple)
                break

    async def replace_session(
            self, basekey: BaseKey, tc: TelegramClient) -> None:
        """
        This method will replace LocalBox session to
        session of specified ``TelegramClient``.

        Arguments:
            basekey (``BaseKey``):
                ``BaseKey`` of this *LocalBox*.

            tc (``TelegramClient``):
                ``TelegramClient`` from which we
                will extract new session.
        """
        try:
            AES(basekey).decrypt(self._elb._session).decode()
        except UnicodeDecodeError:
            raise IncorrectKey(
                'BaseKey doesn\'t match with BaseKey of LocalBox') from None
        else:
            self._session = tc.session.save()

            session = AES(basekey).encrypt(self._session.encode())
            self._elb._session = session

            sql_tuple = ('UPDATE BOX_DATA SET SESSION = ?',(session,))
            await self._tgbox_db.BOX_DATA.execute(sql_tuple)

    async def search_file(
            self, sf: SearchFilter,
            cache_preview: bool=True) -> AsyncGenerator[
                'DecryptedLocalBoxFile', None
            ]:
        """
        This method used to search for files in your ``DecryptedLocalBox``.

        Arguments:
            sf (``SearchFilter``):
                ``SearchFilter`` with kwargs you like.

            cache_preview (``bool``, optional):
                Will cache preview in file object if ``True``.
        """
        async for file in search_generator(
                sf, lb=self, cache_preview=cache_preview):
                    yield file

    async def prepare_file(
            self, file: Union[BinaryIO, bytes, Document, Photo],
            file_size: Optional[int] = None,
            file_path: Optional[Path] = None,
            cattrs: Optional[Dict[str, Union[bytes]]] = None,
            make_preview: bool=True) -> 'PreparedFile':
        """
        Prepares your file for ``RemoteBox.push_file``

        Arguments:
            file (``BinaryIO``, ``BytesIO``):
                ``file`` data to add to the LocalBox. In most
                cases it's just opened file. If you want to upload
                something else, then you need to implement class
                that have ``read`` & ``name`` methods.

                The method needs to know size of the ``file``, so
                it will try to ask system what size of file on path
                ``file.name``. If it's impossible, then method tries to
                seek file to EOF, if file isn't seekable, then we try to
                get size by ``len()`` (as ``__len__`` dunder). If all fails,
                method tries to get ``file.read())`` (with load to RAM).

                Abs file path length must be <= ``self.defaults.FILE_PATH_MAX``;
                If file has no ``name`` and ``file_path`` is not
                specified then it will be ``NO_FOLDER/{prbg(6).hex()}``.

            file_size (``int``, optional):
                Bytelength of ``file``. You can specify
                it if you already know file size.

            file_path (``Path``, optional):
                File path. You can specify here path
                that isn't exists. If not specified, will
                be used path from the ``BinaryIO``, if file
                is not a ``BinaryIO`` then will be used a
                ``self.defaults.DEF_NO_FOLDER``.

                Must be <= ``self.defaults.FILE_PATH_MAX``.

            cattrs (``Dict[str, Union[bytes, None]]``, optional):
                The file's custom metadata attributes that
                will be added to the main metadata. Specified
                dict will be packed with the PackedAttributes.

                Please note that after some operations we
                will create a file metadata. Its limit
                defined as ``self.defaults.METADATA_MAX``. You
                shouldn't overflow this number plus size of
                default metadata; other is up to you.

            make_preview (``bool``, optional):
                Will try to add file preview to
                the metadata if ``True`` (default).
        """
        file_salt, file_iv = get_rnd_bytes(32), get_rnd_bytes(16)
        filekey = make_filekey(self._mainkey, file_salt)

        if isinstance(file, (Document, Photo)):
            if not self._session:
                raise NotEnoughRights(
                    '''You need to decrypt LocalBox with BaseKey, '''
                    '''MainKey is not enough. Session is None.'''
                )
            file = _TelegramVirtualFile(
                file, self._session,
                self._api_id, self._api_hash
            )
            # We will call get_preview
            make_preview = False

        if file_path is None:
            if hasattr(file,'name') and file.name:
                file_path = Path(file.name).absolute()
            else:
                file_path = Path(self._defaults.DEF_NO_FOLDER, prbg(8).hex())
        else:
            if len(file_path.parts) < 2:
                raise ValueError('Path should contain folder and file name')

        if len(str(file_path)) > self._defaults.FILE_PATH_MAX:
            raise LimitExceeded(f'File path must be <= {self._defaults.FILE_PATH_MAX} bytes.')

        file_fingerprint = sha256(
            str(file_path).encode()\
          + self._mainkey.key).digest()

        await self._check_fingerprint(file_fingerprint)

        if not file_size:
            if isinstance(file, _TelegramVirtualFile):
                file_size = file.size
            else:
                try:
                    file_size = getsize(file.name)
                except (FileNotFoundError, AttributeError):
                    if isinstance(file, bytes):
                        file_size = len(file)
                        file = BytesIO(file)

                    elif hasattr(file, '__len__'):
                        file_size = len(file)

                    elif hasattr(file,'seek') and file.seekable():
                        file.seek(0,2)
                        file_size = file.tell()
                        file.seek(0,0)
                    else:
                        rb = file.read()
                        file_size = len(rb)
                        file = BytesIO(rb)
                        del rb

        if file_size <= 0:
            raise InvalidFile('Specified file is empty or file_size in invalid')

        if isinstance(file, _TelegramVirtualFile):
            if file.mime:
                mime_type = file.mime
                file_type = mime_type.split('/')[0]
            else:
                mime_type = ''
                file_type = None
        else:
            try:
                mime_type = filetype_guess(file).mime
                file_type = mime_type.split('/')[0]
            except (TypeError, FileNotFoundError, AttributeError):
                mime_type = ''
                file_type = None
            finally:
                # filetype.guess reads N bytes and
                # doesn't seek back.
                file.seek(0,0)

        preview, duration = b'', 0

        if isinstance(file, _TelegramVirtualFile):
            preview = await file.get_preview()
            duration = file.duration
        else:
            if make_preview and file_type in ('audio','video','image'):
                try:
                    preview = (await make_media_preview(file.name)).read()
                except PreviewImpossible:
                    pass

                if file_type in ('audio','video'):
                    try:
                        duration = await get_media_duration(file.name)
                    except DurationImpossible:
                        pass

        # --- Start constructing metadata here --- #

        # We should always encrypt FILE_PATH with MainKey.
        file_path_no_name = str(file_path.parent).encode()
        efile_path = AES(self._mainkey).encrypt(file_path_no_name)

        cattrs = PackedAttributes.pack(**cattrs) if cattrs else b''

        secret_metadata = PackedAttributes.pack(
            preview = preview,
            duration = int_to_bytes(duration),
            file_size = int_to_bytes(file_size),
            efile_path = efile_path,
            file_name = file_path.name.encode(),
            mime = mime_type.encode(),
            cattrs = cattrs
        )
        secret_metadata = AES(filekey).encrypt(secret_metadata)

        metadata = PackedAttributes.pack(
            box_salt = self._box_salt,
            file_salt = file_salt,
            file_fingerprint = file_fingerprint,
            secret_metadata = secret_metadata
        )
        if len(metadata) > self._defaults.METADATA_MAX:
            raise LimitExceeded(
                f'Total len(metadata) must be <= {self._defaults.METADATA_MAX}'
            )
        if len(metadata) > 256**3-1:
            raise LimitExceeded(
                'Total len(metadata) must be <= 256^3-1'
            )
        metadata_bytesize = int_to_bytes(len(metadata),3)

        constructed_metadata =  PREFIX + VERBYTE
        constructed_metadata += metadata_bytesize
        constructed_metadata += metadata + file_iv

        total_file_size = len(constructed_metadata) + file_size

        return PreparedFile(
            dlb = self,
            file = file,
            filekey = filekey,
            filesize = total_file_size,
            filepath = Path(file_path_no_name.decode()),
            filesalt = file_salt,
            fingerprint = file_fingerprint,
            metadata = constructed_metadata,
            imported = False
        )
    async def import_file(
            self, drbf: 'tgbox.api.remote.DecryptedRemoteBoxFile',
            file_path: Optional[Union[str, Path]] = None)\
            -> 'DecryptedLocalBoxFile':
        """
        Imports file to your ``DecryptedLocalBox``

        Arguments:
            drbf (``DecryptedRemoteBoxFile``):
                Remote file you want to import.

            file_path (``Path``, optional):
                File's path. Will be used ``drbf._file_path`` if
                ``None`` and if drbf was decrypted with the
                ``MainKey``, otherwise ``self.defaults.DEF_NO_FOLDER``.

                This method will call a ``set_file_path(file_path)``
                on the specified drbf if ``file_path`` isn't a ``None``.

                You can change drbf ``file_path`` with
                ``set_file_path`` method before importing
                file, so you don't need to specify it here.
        """
        # We need to fetch encrypted metadata
        if not drbf._erbf._initialized:
            await drbf._erbf.init()

        if not file_path:
            if drbf.file_path:
                file_path = drbf.file_path
            else:
                file_path = self._defaults.DEF_NO_FOLDER

        if isinstance(file_path, str):
            file_path = Path(file_path)

        drbf.set_file_path(file_path)

        pf = PreparedFile(
            dlb = self,
            file = BytesIO(),
            filekey = drbf._filekey,
            filesize = drbf._size,
            filepath = file_path,
            filesalt = drbf._file_salt,
            fingerprint = drbf._fingerprint,
            metadata = drbf._erbf._metadata,
            imported = True
        )
        pf.set_file_id(drbf._id)
        pf.set_upload_time(drbf._upload_time)

        return await self._make_local_file(pf)

    def get_sharekey(self, reqkey: Optional[RequestKey] = None) -> ShareKey:
        """
        Returns ``ShareKey`` for this Box. You should use
        this method if you want to share your LocalBox
        with other people.

        Arguments:
            reqkey (``RequestKey``, optional):
                Other's ``RequestKey``. If isn't specified
                returns ``ImportKey`` of this box without
                encryption, so anyone with this key can
                decrypt **ALL** files in your Boxes.
        """
        if reqkey:
            return make_sharekey(
                requestkey=reqkey,
                mainkey=self._mainkey,
                box_salt=self._box_salt
            )
        else:
            return make_sharekey(mainkey=self._mainkey)

class EncryptedLocalBoxDirectory:
    """
    Class that represents abstract tgbox directory. You
    can iterate over all files/folders in it, as well
    as load parent folder up to root.

    Typical usage:

    .. code-block:: python

        from asyncio import run as asyncio_run
        from tgbox.api import get_localbox
        from tgbox.keys import make_basekey, Phrase

        phrase = Phrase(b'example phrase here')
        basekey = make_basekey(phrase)

        async def main():
            dlb = await get_localbox(basekey)
            dlbfi = await dlb.get_file(await dlb.get_last_file_id())

            # Load directory up to the root
            await dlbfi.directory.lload(full=True)

            print(dlbfi.directory)

            # Iterate over files/folders in this directory
            async for content in dlbfi.directory.iterdir():
                print(content) # May be file or another dir

        asyncio_run(main())
    """
    def __init__(self, tgbox_db: TgboxDB, part_id: bytes):
        """
        Arguments:
            tgbox_db (``TgboxDB``):
                Tgbox Database.

            part_id (``bytes``):
                Path's part ID. You can fetch it from
                the ``PATH_PARTS`` table in ``TgboxDB``.
        """
        self._initialized = False
        self._tgbox_db = tgbox_db
        self._lb = EncryptedLocalBox(self._tgbox_db)

        self._part = None
        self._part_id = part_id
        self._parent_part_id = None

        self._parts = [self]
        self._floaded = False

    def __hash__(self) -> int:
        x = tuple(i.part for i in self._parts)
        # Without 22 hash of list will be equal to object's
        return hash((x, 22))

    def __eq__(self, other) -> bool:
        return all((
            isinstance(other, self.__class__),
            self.__hash__() == hash(other)
        ))
    def __str__(self) -> str:
        """Will return path for current loaded parts"""
        if isinstance(self._lb, DecryptedLocalBox):
            return str(Path(*[i.part.decode() for i in self._parts]))
        else:
            return self.__repr__()

    def __repr__(self) -> str:
        c = 'ELBD' if 'Encrypted' in self.__class__.__name__ else 'DLBD'
        return f'{c}[{self.part.decode() if c == "DLBD" else self.part}]'

    def __getitem__(self, _slice: slice):
        return self.parts[_slice]

    def __raise_initialized(self) -> NoReturn:
        if not self._initialized:
            raise NotInitializedError('Not initialized. Call .init().')

    @property
    def initialized(self) -> bool:
        """
        Returns ``True`` if you
        already called ``.init()``
        """
        return self._initialized

    @property
    def part(self) -> Union[bytes, None]:
        """
        Returns ``None`` if you didn't called
        the ``.init()`` method, encrypted path part
        if you're in ``EncryptedLocalBoxDirectory``,
        and decrypted if ``DecryptedLocalBoxDirectory``.
        """
        return self._part

    @property
    def part_id(self) -> bytes:
        """Returns path part ID"""
        return self._part_id

    @property
    def parent_part_id(self) -> Union[bytes, None]:
        """
        Returns ``None`` if you didn't called the ``.init()``
        method, parent path part ID otherwise.
        """
        return self._parent_part_id

    @property
    def parts(self) -> list:
        """Will return loaded path parts."""
        return self._parts

    @property
    def floaded(self) -> bool:
        """Will return ``True`` if path was fully loaded"""
        return self._floaded

    async def init(self) -> 'EncryptedLocalBoxDirectory':
        """Will fetch required data from the database."""
        folder_row = await self._tgbox_db.PATH_PARTS.select_once((
            'SELECT * FROM PATH_PARTS WHERE PART_ID=?',
            (self._part_id,)
        ))
        self._part = folder_row[0]
        self._part_id = folder_row[1]
        self._parent_part_id = folder_row[2]

        if not self._lb.initialized:
            await self._lb.init()

        self._initialized = True
        return self

    async def lload(self, full: Optional[bool] = False):
        """
        Will load and return one path part
        from the left side (previous) or
        return ``None`` if there is nothing left.

        Arguments:
            full (``bool``, optional):
                If ``full`` is ``True``, will load all parts
                from the left side and return ``None``.
        """
        self.__raise_initialized()

        for _ in range(256**8 if full else 1):
            previous_part = await self._tgbox_db.PATH_PARTS.select_once((
                'SELECT PARENT_PART_ID FROM PATH_PARTS WHERE PART_ID=?',
                (self.parts[0].part_id,)
            ))
            if not previous_part[0]:
                self._floaded = True
                return

            if isinstance(self._lb, DecryptedLocalBox):
                previous_part = await EncryptedLocalBoxDirectory(
                    self._tgbox_db, previous_part[0]).decrypt(self._lb._mainkey)
            else:
                previous_part = await EncryptedLocalBoxDirectory(
                    self._tgbox_db, previous_part[0]).init()

            self.parts.insert(0, previous_part)

        return previous_part

    async def iterdir(
        self,
        ignore_dirs: bool=False,
        ignore_files: bool=False,
        cache_preview: bool=True,
        ppid: Optional[Union[bytes, DirectoryRoot]] = None) -> Union[
            'EncryptedLocalBoxFile',
            'DecryptedLocalBoxFile',
            'EncryptedLocalBoxDirectory',
            'DecryptedLocalBoxDirectory'
        ]:
        """
        Iterate over all files/folders
        in this abstract directory.

        Arguments:
            ignore_dirs (``bool``, optional):
                Return abstract folders from this
                directory as LocalBoxDirectory or
                not. If ``False`` will return only
                LocalBoxFile objects.

            ignore_files (``bool``, optional):
                Will return LocalBoxFile if ``False``.

            cache_preview (``bool``, optional):
                Cache preview in class or not.
                ``True`` by default.

            ppid (``bytes``, ``DirectoryRoot``):
                Path PartID to iterate in. Will iterate over
                absolute LocalBox directory root if it's
                ``DirectoryRoot``. Will use ``self.part_id``
                if not specified (by default).
        """
        assert not all((ignore_files, ignore_dirs)), 'Specify at least one'

        if isinstance(ppid, DirectoryRoot) or ppid is DirectoryRoot:
            part_id = None
        elif ppid:
            part_id = ppid
        else:
            part_id = self._part_id

        if not ignore_dirs:
            folders = await self._tgbox_db.PATH_PARTS.execute((
                'SELECT * FROM PATH_PARTS WHERE PARENT_PART_ID IS ?',
                (part_id,)
            ))
            async for folder_row in folders:
                if isinstance(self._lb, DecryptedLocalBox):
                    yield await EncryptedLocalBoxDirectory(self._tgbox_db,
                        folder_row[1]).decrypt(self._lb._mainkey)
                else:
                    yield await EncryptedLocalBoxDirectory(self._tgbox_db,
                        folder_row[1]).init()

        if not ignore_files:
            files = await self._tgbox_db.FILES.execute((
                'SELECT * FROM FILES WHERE PPATH_HEAD IS ?',
                (part_id,)
            ))
            async for file_row in files:
                yield await self._lb.get_file(
                    file_row[0], cache_preview=cache_preview)

    async def get_files_total(self) -> int:
        """Will return a total number of files in this directory"""
        cursor = await self._tgbox_db.FILES.execute((
            'SELECT count(*) FROM FILES WHERE PPATH_HEAD IS ?',
            (self.part_id,))
        )
        return (await cursor.fetchone())[0]

    async def get_folders_total(self) -> int:
        """Will return a total number of folders in this directory"""
        cursor = await self._tgbox_db.PATH_PARTS.execute((
            'SELECT count(*) FROM PATH_PARTS WHERE PARENT_PART_ID IS ?',
            (self.part_id,))
        )
        return (await cursor.fetchone())[0]

    async def get_contents_total(self) -> int:
        """Will return a total number of contents in this directory"""
        total_files = await self.get_files_total()
        total_folders = await self.get_folders_total()

        return total_files + total_folders

    async def delete(self) -> None:
        """
        Will delete this folder with all files from your LocalBox.
        All files will stay in ``RemoteBox``, so you can restore
        all your folders by importing files.
        """
        await self._tgbox_db.FILES.execute(
            ('DELETE FROM FILES WHERE PPATH_HEAD=?',(self._part_id,))
        )
        await self._tgbox_db.PATH_PARTS.execute(
            ('DELETE FROM PATH_PARTS WHERE PART_ID=?',(self._part_id,))
        )
    async def decrypt(self, key: Union[BaseKey, MainKey]):
        """Decrypt self and return ``DecryptedLocalBoxDirectory``"""
        if not self._initialized: await self.init()
        return DecryptedLocalBoxDirectory(self, key)

class DecryptedLocalBoxDirectory(EncryptedLocalBoxDirectory):
    def __init__(
            self, elbd: EncryptedLocalBoxDirectory,
            key: Union[BaseKey, MainKey]):
        """
        Arguments:
            elbd (``EncryptedLocalBoxDirectory``):
                Initialized ``EncryptedLocalBoxDirectory``.

            key (``BaseKey``, ``MainKey``):
                Decryption ``Key``.
        """
        super().__init__(elbd._tgbox_db, elbd._part_id)

        self._initialized = True
        self._elbd = elbd

        self._lb = DecryptedLocalBox(elbd._lb, key)
        self._part = AES(self._lb._mainkey).decrypt(elbd._part)

    @staticmethod
    async def init() -> NoReturn:
        raise AttributeError(
            """This function was inherited from ``EncryptedLocalBoxDirectory`` """
            """and cannot be used on ``DecryptedLocalBoxDirectory``."""
        )
    @staticmethod
    async def decrypt() -> NoReturn:
        raise AttributeError(
            """This function was inherited from ``EncryptedLocalBoxDirectory`` """
            """and cannot be used on ``DecryptedLocalBoxDirectory``."""
        )
class EncryptedLocalBoxFile:
    """
    This class represents an encrypted local file. On
    more low-level that's a wrapper around row of
    ``FILES`` table in Tgbox Database. Usually you
    will not use this in your code.

    .. code-block:: python

        from asyncio import run as asyncio_run
        from tgbox.api import get_localbox
        from tgbox.keys import make_basekey, Phrase

        phrase = Phrase(b'example phrase here')
        basekey = make_basekey(phrase)

        async def main():
            dlb = await get_localbox(basekey)

            elbf = await dlb.get_file(
                id = await dlb.get_last_file_id(),
                decrypt = False
            )
            print(elbf.file_salt.hex())
            print(elbf.box_salt.hex())

        asyncio_run(main())
    """
    def __init__(
            self, id: int, tgbox_db: TgboxDB,
            cache_preview: bool=True,
            defaults: Optional[Union[DefaultsTableWrapper,
                RemoteBoxDefaults]] = None) -> None:
        """
        Arguments:
            id (``int``):
                File ID.

            tgbox_db (``TgboxDB``):
                Tgbox Database.

            cache_preview (``bool``, optional):
                Cache preview in class or not.

            defaults (``DefaultsTableWrapper``, ``RemoteBoxDefaults``):
                Class with a default values/constants we will use.
        """
        self._id = id
        self._tgbox_db = tgbox_db
        self._cache_preview = cache_preview

        self._initialized = False
        self._fingerprint = None
        self._updated_metadata = None

        self._ppath_head, self._upload_time = None, None
        self._metadata, self._directory = None, None
        self._imported, self._efilekey = None, None
        self._file_salt, self._version_byte = None, None

        if defaults:
            self._defaults = defaults
        else:
            self._defaults = DefaultsTableWrapper(self._tgbox_db)

    def __hash__(self) -> int:
        return hash((self._id, 22))

    def __eq__(self, other) -> bool:
        return all((
            isinstance(other, self.__class__),
            self.__hash__() == hash(other)
        ))
    @property
    def initialized(self) -> bool:
        """
        Returns ``True`` if you
        already called ``.init()``
        """
        return self._initialized

    @property
    def imported(self) -> Union[bool, None]:
        """
        Returns ``True`` if file was
        forwarded to your BoxChannel.
        """
        return self._imported

    @property
    def fingerprint(self) -> Union[bytes, None]:
        """
        Returns file fingerprint (hash of
        file path plus mainkey) or ``None``
        """
        return self._fingerprint

    @property
    def version_byte(self) -> Union[bytes, None]:
        """Returns Verbyte of this file or
        ``None`` if class wasn't initialized
        """
        return self._version_byte

    @property
    def defaults(self) -> Union[DefaultsTableWrapper, RemoteBoxDefaults]:
        """
        Will return ``DefaultsTableWrapper`` or
        ``RemoteBoxDefaults``.
        """
        return self._defaults

    @property
    def directory(self) -> Union[
            EncryptedLocalBoxDirectory,
            DecryptedLocalBoxDirectory,
            None
        ]:
        """
        Returns ``None`` i you call it on
        ``EncryptedLocalBoxFile`` that wasn't
        initialized, ``EncryptedLocalBoxDirectory``
        if on initialized or ``DecryptedLocalBoxDirectory``
        if you call it on ``DecryptedLocalBoxFile``
        """
        return self._directory

    @property
    def id(self) -> Union[int, None]:
        """
        Returns file ID or ``None``
        if file wasn't initialized
        """
        return self._id

    @property
    def upload_time(self) -> Union[bytes, int, None]:
        """
        Returns encrypted ``upload_time`` from
        ``EncryptedLocalBoxFile`` and decrypted
        from ``DecryptedLocalBoxFile``. ``None``
        if class wasn't initialized.
        """
        return self._upload_time

    @property
    def file_iv(self) -> Union[bytes, None]:
        """
        Returns file IV or ``None``
        if file wasn't initialized
        """
        return self._file_iv

    @property
    def file_salt(self) -> Union[bytes, None]:
        """
        Returns file salt or ``None``
        if file wasn't initialized
        """
        return self._file_salt

    @property
    def box_salt(self) -> Union[bytes, None]:
        """
        Returns box salt or ``None``
        if file wasn't initialized
        """
        return self._box_salt

    @property
    def prefix(self) -> Union[bytes, None]:
        """
        Returns file prefix or ``None``
        if file wasn't initialized
        """
        return self._prefix

    def __raise_initialized(self) -> NoReturn:
        if not self._initialized:
            raise NotInitializedError('Not initialized. Call .init().')

    async def init(self) -> 'EncryptedLocalBoxFile':
        """Will fetch and parse data from the Database."""

        file_row = list(await self._tgbox_db.FILES.select_once(
            sql_tuple = ('SELECT * FROM FILES WHERE ID=?', (self._id,))
        ))
        self._updated_metadata = file_row.pop()
        self._metadata = file_row.pop()
        self._fingerprint = file_row.pop()
        self._efilekey = file_row.pop()
        self._ppath_head = file_row.pop()
        self._upload_time = file_row.pop()
        self._id = file_row.pop()

        self._directory = EncryptedLocalBoxDirectory(
            self._tgbox_db, self._ppath_head
        )
        await self._directory.init()

        self._imported = bool(self._efilekey)

        self._prefix = self._metadata[:len(PREFIX)]
        self._version_byte = self._metadata[
            len(PREFIX) : len(VERBYTE) + len(PREFIX)
        ]
        pattr_offset = len(PREFIX) + len(VERBYTE) + 3

        unpacked_metadata = PackedAttributes.unpack(
            self._metadata[pattr_offset:-16]
        )
        self._file_iv = self._metadata[-16:]
        self._file_salt = unpacked_metadata['file_salt']
        self._box_salt = unpacked_metadata['box_salt']

        if isinstance(self._defaults, DefaultsTableWrapper):
            if not self._defaults.initialized:
                await self._defaults.init()

        self._initialized = True
        return self

    def disable_cache_preview(self) -> None:
        """
        Sets ``self._cache_preview`` to ``False``
        and removes cached preview from memory.
        """
        self._cache_preview = False
        self._preview = b''

    def enable_cache_preview(self) -> None:
        """
        Sets ``self._cache_preview`` to ``True``.
        Preview will be cached after first
        ``object.preview`` call.
        """
        self._cache_preview = True

    async def decrypt(self, key: Union[FileKey, MainKey]) -> 'DecryptedLocalBoxFile':
        """
        Returns decrypted by ``key`` ``EncryptedLocalBoxFile``

        Arguments:
            key (``FileKey``, ``MainKey``):
                Decryption key.
        """
        if not self.initialized:
            await self.init()
        return DecryptedLocalBoxFile(self, key)

    async def delete(self) -> None:
        """
        Will delete this file from your LocalBox. You can
        re-import it from ``RemoteBox`` with ``import_file``.

        .. note::
            This will delete file only from your LocalBox.
            To completly remove your file use same
            function on ``EncryptedRemoteBoxFile``.
        """
        # Getting file row for PPATH_HEAD
        file_row = await self._tgbox_db.FILES.select_once(
            sql_tuple=('SELECT PPATH_HEAD FROM FILES WHERE ID=?',(self._id,))
        )
        # Removing requested file
        await self._tgbox_db.FILES.execute(
            ('DELETE FROM FILES WHERE ID=?',(self._id,))
        )
        try:
            await self._tgbox_db.FILES.select_once(sql_tuple=(
                'SELECT ID FROM FILES WHERE PPATH_HEAD=?',(file_row[0],)
            ))
        # Only one file point to this path part (folder)
        except StopAsyncIteration:
            await self._tgbox_db.PATH_PARTS.execute((
                'DELETE FROM PATH_PARTS WHERE PART_ID=?',(file_row[0],)
            ))

    def get_requestkey(self, mainkey: MainKey) -> RequestKey:
        """
        Returns ``RequestKey`` for this File. You
        should use this method if you want to decrypt
        other's ``EncryptedLocalBoxFile``.

        Arguments:
            mainkey (``MainKey``):
                To make a ``RequestKey`` for other's
                ``EncryptedLocalBoxFile`` you need to have
                your own Box. Take key from it and specify here.
        """
        self.__raise_initialized()
        return make_requestkey(mainkey, file_salt=self._file_salt)

class DecryptedLocalBoxFile(EncryptedLocalBoxFile):
    """
    This class represents an decrypted local file.
    On more low-level that's a wrapper of ``FILES``
    table in Tgbox Database that decrypts row.

    Typical usage:

    .. code-block:: python

        from asyncio import run as asyncio_run
        from tgbox.api import get_localbox
        from tgbox.keys import make_basekey, Phrase

        phrase = Phrase(b'example phrase here')
        basekey = make_basekey(phrase)

        async def main():
            dlb = await get_localbox(basekey)
            lfid = dlb.get_last_file_id()
            dlbfi = await dlb.get_file(lfid)

            print(dlbfi.id, dlbfi.id, dlbfi.size)

        asyncio_run(main())
    """
    def __init__(
            self, elbf: EncryptedLocalBoxFile,
            key: Union[FileKey, ImportKey, MainKey],
            cache_preview: bool=True):
        """
        Arguments:
            elbf (``EncryptedLocalBoxFile``):
                Encrypted local box file that
                you want to decrypt.

            key (``FileKey``, ``ImportKey``, ``MainKey``):
                Decryption key.

            cache_preview (``bool``, optional):
                Cache preview in class or not.
        """
        if not elbf._initialized:
            raise NotInitializedError('You should init elbf firstly')

        self._initialized = True

        self._elbf = elbf
        self._key = key
        self._tgbox_db = elbf._tgbox_db

        self._efilekey = elbf._efilekey

        self._id = elbf._id
        self._imported = elbf._imported
        self._ppath_head = elbf._ppath_head

        self._fingerprint = elbf._fingerprint
        self._updated_metadata = elbf._updated_metadata
        self._defaults = elbf._defaults

        self._prefix = elbf._prefix
        self._version_byte = elbf._version_byte
        self._file_iv = elbf._file_iv
        self._file_salt = elbf._file_salt
        self._box_salt = elbf._box_salt

        if cache_preview is None:
            self._cache_preview = elbf._cache_preview
        else:
            self._cache_preview = cache_preview

        if isinstance(key, (FileKey, ImportKey)):
            self._filekey = FileKey(key.key)
        elif isinstance(key, MainKey) and self._efilekey:
            self._filekey = FileKey(
                AES(self._key).decrypt(self._efilekey)
            )
        else:
            self._filekey = make_filekey(self._key, self._file_salt)

        if isinstance(key, MainKey):
            self._mainkey = key
        else:
            self._mainkey = None

        self._upload_time = AES(self._filekey).decrypt(elbf._upload_time)
        self._upload_time = bytes_to_int(self._upload_time)

        pattr_offset = len(PREFIX) + len(VERBYTE) + 3

        unpacked_metadata = PackedAttributes.unpack(
            self._elbf._metadata[pattr_offset:-16]
        )
        secret_metadata = AES(self._filekey).decrypt(
            unpacked_metadata['secret_metadata']
        )
        if len(secret_metadata) == len(unpacked_metadata['secret_metadata'])-16:
            raise AESError('Metadata wasn\'t decrypted correctly. Incorrect key?')

        secret_metadata = PackedAttributes.unpack(secret_metadata)

        self.__required_metadata = [
            'duration', 'file_size', 'file_name',
            'cattrs', 'mime', 'efile_path', 'preview'
        ]
        for attr in self.__required_metadata:
            setattr(self, f'_{attr}', secret_metadata[attr])

        self._size = bytes_to_int(self._file_size) # pylint: disable=no-member
        del self._file_size # pylint: disable=no-member

        self._duration = bytes_to_int(self._duration)
        self._cattrs = PackedAttributes.unpack(self._cattrs)
        self._mime = self._mime.decode()

        self._file_name = self._file_name.decode()

        if not cache_preview:
            self._preview = b''

        if self._mainkey and not self._efilekey:
            self._file_path = AES(self._mainkey).decrypt(
                secret_metadata['efile_path']
            )
            self._file_path = Path(self._file_path.decode())
        else:
            self._file_path = self._defaults.DEF_NO_FOLDER

        for attr in self.__required_metadata:
            secret_metadata.pop(attr)

        self._residual_metadata = secret_metadata

        if self._mainkey:
            self._directory = DecryptedLocalBoxDirectory(
                self._elbf._directory, self._mainkey
            )
        else:
            self._directory = self._elbf._directory

        self._download_path = self._defaults.DOWNLOAD_PATH

        if self._elbf._updated_metadata:
            updates = AES(self._filekey).decrypt(
                self._elbf._updated_metadata
            )
            updates = PackedAttributes.unpack(updates)

            for k,v in tuple(updates.items()):
                if k in self.__required_metadata:
                    if k == 'cattrs':
                        setattr(self, f'_{k}', PackedAttributes.unpack(v))

                    elif k == 'file_name':
                        setattr(self, f'_{k}', v.decode())

                    elif k == 'efile_path':
                        if self._mainkey and not self._efilekey:
                            self._file_path = AES(self._mainkey).decrypt(v)
                            self._file_path = Path(self._file_path.decode())
                        else:
                            self._file_path = self._defaults.DEF_NO_FOLDER
                    else:
                        setattr(self, f'_{k}', v)
                else:
                    self._residual_metadata[k] = v

        self._elbf._initialized = False
        self._elbf._metadata = None # To save RAM
        self._elbf._updated_metadata = None

    @staticmethod
    async def init() -> NoReturn:
        raise AttributeError(
            """This function was inherited from ``EncryptedLocalBoxFile`` """
            """and cannot be used on ``DecryptedLocalBoxFile``."""
        )
    @staticmethod
    async def decrypt() -> NoReturn:
        raise AttributeError(
            """This function was inherited from ``EncryptedLocalBoxFile`` """
            """and cannot be used on ``DecryptedLocalBoxFile``."""
        )
    @property
    def residual_metadata(self) -> dict:
        """
        Will return metadata that left after
        parsing secret_metadata. This can be
        useful in future, when lower version
        will read file of a higher version.
        """
        return self._residual_metadata

    @property
    def file_path(self) -> Path:
        """Returns file path."""
        return self._file_path

    @property
    def file_name(self) -> str:
        """Returns file name."""
        return self._file_name

    @property
    def preview(self) -> Union[bytes, None]:
        """
        Returns preview bytes or ``b''``
        if ``cache_preview`` is ``False``.
        """
        return self._preview

    @property
    def size(self) -> int:
        """Returns file size (no metadata included)."""
        return self._size

    @property
    def duration(self) -> int:
        """Returns media file duration."""
        return self._duration

    @property
    def cattrs(self) -> Union[bytes, None]:
        """Returns file Custom Attributes"""
        return self._cattrs

    @property
    def mime(self) -> Union[bytes, None]:
        """Returns mime type of the file"""
        return self._mime

    @property
    def download_path(self) -> Path:
        """Returns current download path"""
        return self._download_path

    def set_download_path(self, path: Path):
        """Will set download path to specified."""
        self._download_path = path

    async def refresh_metadata(
            self, drb: Optional['tgbox.api.remote.DecryptedRemoteBox'] = None,
            _updated_metadata: Optional[bytes] = None
        ):
        """
        This method will refresh local UPDATED_METADATA from
        the remote box file. You should call it after
        every ``DecryptedRemoteBoxFile.update_metadata``
        await, or specify ``DecryptedLocalBox`` when
        awaiting ``update_metadata`` as ``dlb`` kwarg.

        Arguments:
            drb (``DecryptedRemoteBox``, optional):
                ``DecryptedRemoteBox`` associated with
                this ``DecryptedLocalBox``.

            _updated_metadata (``bytes``, optional):
                Updated metadata by itself. This is for
                internal use, specify only ``drb``.

        You should specify at least one argument.
        """
        assert any((drb, _updated_metadata)), 'Specify at least one'

        if not _updated_metadata:
            drbf = await drb.get_file(self._id)
            _updated_metadata = drbf._message.message

        await self._tgbox_db.FILES.execute((
            'UPDATE FILES SET UPDATED_METADATA=? WHERE ID=?',
            (_updated_metadata, self._id)
        ))
    def get_sharekey(self, reqkey: Optional[RequestKey] = None) -> ShareKey:
        """
        Returns ``ShareKey`` for this file. You should
        use this method if you want to share your
        ``DecryptedLocalBoxFile`` with other people.

        Arguments:
            reqkey (``RequestKey``, optional):
                Other's ``RequestKey``. If isn't specified
                returns ``ShareKey`` of this file without
                encryption, so ANYONE with this key can
                decrypt this local & remote box file.
        """
        if reqkey:
            return make_sharekey(
                requestkey=reqkey, filekey=self._filekey,
                file_salt=self._file_salt
            )
        else:
            return make_sharekey(filekey=self._filekey)



"""This module stores all Tgbox-unique exceptions."""

class TgboxException(Exception):
    """Base TGBOX Exception"""

# Base Exceptions

class NotInitializedError(TgboxException):
    """The class you try to use isn't initialized"""

class PathIsDirectory(TgboxException):
    """Specified path is Directory"""

class LimitExceeded(TgboxException):
    """Value is out of allowed range"""

class NotATgboxFile(TgboxException):
    """This Telegram message isn't a TGBOX file"""

class InvalidFile(TgboxException):
    """Specified file is invalid"""

# Crypto Exceptions

class IncorrectKey(TgboxException):
    """Specified key is invalid."""

class ModeInvalid(TgboxException):
    """You should use only decryption or encryption per class"""

class AESError(TgboxException):
    """Invalid AES configuration"""

# Tools Exceptions

class ConcatError(TgboxException):
    """You must concat metadata before using OpenPretender"""

class PreviewImpossible(TgboxException):
    """Can\'t create file preview"""

class DurationImpossible(TgboxException):
    """Can\'t get media duration"""

# Database Exceptions

class InUseException(TgboxException):
    """The DB already exists and in use"""

class BrokenDatabase(TgboxException):
    """Can\'t parse SQLite DB"""

# RemoteBox Exceptions

class RemoteFileNotFound(TgboxException):
    """Seems that there is no requested by you file"""

class SessionUnregistered(TgboxException):
    """Session you trying to use was disconnected"""

class RemoteBoxInaccessible(TgboxException):
    """The RemoteBox you try to use is inaccessible"""

class NotEnoughRights(TgboxException):
    """You don't have rigths for this action"""

class NoPlaceLeftForMetadata(TgboxException):
    """Your edited metadata overflow Telegram caption limit"""

# LocalBox Exceptions

class AlreadyImported(TgboxException):
    """LocalBox have file with same ID"""

class NotImported(TgboxException):
    """The file you try to retrieve wasn\'t imported yet"""

class FingerprintExists(TgboxException):
    """File with the same file path already uploaded to the Box"""
